import torch


def get_num_zero_dimensions():
    # while one dimension would suffice, using only one can cause (especially when using
    # single_targets=True) that the batcher by pytorch removes the singleton dimension. specifically,
    # with only one target per batch item, the size of the knowledge source will be 1, 150, 1, whereas the last dimension
    # is removed during batching by pytorch. so to keep things simple, we just add a two here to avoid having a
    # singleton dimension at the end
    return 2


def get_zeros_as_tensor(term: str):
    tensor_emotions = torch.zeros(get_num_zero_dimensions(), dtype=torch.long)
    return tensor_emotions



def _parse_categories(lines):
    """
    Read (category_id, category_name) pairs from the categories section.
    Each line consists of an integer followed a tab and then the category name.
    This section is separated from the lexicon by a line consisting of a single "%".
    """
    for line in lines:
        line = line.strip()
        if line == "%":
            return
        # ignore non-matching groups of categories
        if "\t" in line:
            category_id, category_name = line.split("\t", 1)
            yield category_id, category_name


def _parse_lexicon(lines, category_mapping):
    """
    Read (match_expression, category_names) pairs from the lexicon section.
    Each line consists of a match expression followed by a tab and then one or more
    tab-separated integers, which are mapped to category names using `category_mapping`.
    """
    for line in lines:
        line = line.strip()
        parts = line.split("\t")
        yield parts[0], [category_mapping[category_id] for category_id in parts[1:]]


def read_dic(filepath):
    """
    Reads a LIWC lexicon from a file in the .dic format, returning a tuple of
    (lexicon, category_names), where:
    * `lexicon` is a dict mapping string patterns to lists of category names
    * `category_names` is a list of category names (as strings)
    """
    with open(filepath) as lines:
        # read up to first "%" (should be very first line of file)
        for line in lines:
            if line.strip() == "%":
                break
        # read categories (a mapping from integer string to category name)
        category_mapping = dict(_parse_categories(lines))
        # read lexicon (a mapping from matching string to a list of category names)
        lexicon = dict(_parse_lexicon(lines, category_mapping))
    return lexicon, list(category_mapping.values())



from argparse import Namespace
from typing import Dict

import torch
import torch.nn as nn

from NewsSentiment.consts import *
from NewsSentiment.dataset import FXDataset
from NewsSentiment.models.FXBaseModel import FXBaseModel


class RandomMulti(FXBaseModel):
    """

    """

    @staticmethod
    def get_language_models():
        """
        All architecture assumes that at least one model is used so we just require
        bert here for compatibility.
        :return:
        """
        return (BERT_BASE_UNCASED,)

    @staticmethod
    def get_input_field_ids():
        return [
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS),
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK),
        ]

    def __init__(self, transformer_models: Dict, opt: Namespace):
        super(RandomMulti, self).__init__()
        self.num_classes = opt.polarities_dim

    def forward(self, inputs):
        text_bert_indices = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS
        )
        batch_size = text_bert_indices.shape[0]
        num_targets = text_bert_indices.shape[1]

        # get a random tensor
        logits = torch.rand(batch_size, num_targets, self.num_classes)

        return logits



import argparse
import math
import os
import random
import sys
import time
from collections import Counter
from typing import Iterable

import numpy
import torch
import torch.nn as nn
from jsonlines import jsonlines
from torch.utils.data import DataLoader
from transformers import (
    BertModel,
    XLNetModel,
    RobertaModel,
    PreTrainedModel,
    BertTokenizer,
    RobertaTokenizer,
    XLNetTokenizer,
    AlbertTokenizer,
    AlbertModel,
)

from NewsSentiment.SentimentClasses import SentimentClasses
from NewsSentiment.consts import *
from NewsSentiment.dataset import FXDataset, RandomOversampler, FXEasyTokenizer
from NewsSentiment.download import Download
from NewsSentiment.earlystopping import EarlyStopping
from NewsSentiment.evaluator import Evaluator
from NewsSentiment.fxlogger import get_logger
from NewsSentiment.knowledge.bingliuopinion.bingliuopinion import get_num_bingliu_polarities
from NewsSentiment.knowledge.mpqasubjectivity.mpqasubjectivity import (
    get_num_mpqa_subjectivity_polarities,
)
from NewsSentiment.knowledge.nrcemolex.nrcemolex import get_num_nrc_emotions
from NewsSentiment.knowledge.zeros.zerosknowledge import get_num_zero_dimensions
from NewsSentiment.losses.crossentropycrossweight import CrossEntropyLoss_CrossWeight
from NewsSentiment.losses.crossentropylosslsr import CrossEntropyLoss_LSR
from NewsSentiment.losses.seq2seqloss import SequenceLoss
from NewsSentiment.models.FXEnsemble import FXEnsemble
from NewsSentiment.models.multitargets.contrasting import Contrasting
from NewsSentiment.models.multitargets.random_multi import RandomMulti
from NewsSentiment.models.multitargets.seq2seq import SeqTwoSeq
from NewsSentiment.models.multitargets.seq2seq_without_targetmask import SeqTwoSeqWithoutTargetMask
from NewsSentiment.models.multitargets.tdbertlikemultitarget import TDBertLikeMultiTarget
from NewsSentiment.models.multitargets.tdbertlikemultitarget_dense import TDBertLikeMultiTargetDense
from NewsSentiment.models.singletarget.aen import AEN_Base
from NewsSentiment.models.singletarget.grutscsingle import GRUTSCSingle
from NewsSentiment.models.singletarget.lcf import LCF_BERT
from NewsSentiment.models.singletarget.lcf2 import LCF_BERT2Dual
from NewsSentiment.models.singletarget.lcfs import LCFS_BERT
from NewsSentiment.models.singletarget.lcfst import LCFST_BERT
from NewsSentiment.models.singletarget.lcft import LCFT_BERT
from NewsSentiment.models.singletarget.notargetcls import NoTargetClsBert
from NewsSentiment.models.singletarget.random_single import RandomSingle
from NewsSentiment.models.singletarget.spc import SPC_Base
from NewsSentiment.models.singletarget.td_bert import TD_BERT
from NewsSentiment.models.singletarget.td_bert_qa import TD_BERT_QA_MUL, TD_BERT_QA_CON
from NewsSentiment.models.singletarget.tdbertlikesingle import TDBertLikeSingle
from NewsSentiment.plotter_utils import create_save_plotted_confusion_matrix

logger = get_logger()

TRANSFORMER_MODELS_INFO = {
    BERT_BASE_UNCASED: {"tokenizer_class": BertTokenizer, "model_class": BertModel},
    ROBERTA_BASE: {"tokenizer_class": RobertaTokenizer, "model_class": RobertaModel},
    XLNET_BASE_CASED: {"tokenizer_class": XLNetTokenizer, "model_class": XLNetModel},
    ALBERT_XXLARGE: {"tokenizer_class": AlbertTokenizer, "model_class": AlbertModel},
    ALBERT_BASE: {"tokenizer_class": AlbertTokenizer, "model_class": AlbertModel},
}

OWN_MODELNAME2CLASS = {
    # class output
    "lcf_bert": LCF_BERT,
    "lcf_bert2": LCF_BERT2Dual,
    "lcfs_bert": LCFS_BERT,
    "lcft_bert": LCFT_BERT,
    "lcfst_bert": LCFST_BERT,
    "spc_bert": SPC_Base,
    "aen_bert": AEN_Base,
    "tdbert": TD_BERT,
    "tdbert-qa-mul": TD_BERT_QA_MUL,
    "tdbert-qa-con": TD_BERT_QA_CON,
    "tdbertlikesingle": TDBertLikeSingle,
    "tdbertlikemulti": TDBertLikeMultiTarget,
    "tdbertlikemulti_dense": TDBertLikeMultiTargetDense,
    "contrasting": Contrasting,
    "grutsc": GRUTSCSingle,
    # sequence output
    "seq2seq": SeqTwoSeq,
    "seq2seq_withouttargetmask": SeqTwoSeqWithoutTargetMask,
    # other baselines
    "notargetclsbert": NoTargetClsBert,
    "random_multi": RandomMulti,
    "random_single": RandomSingle,
    # misc
    "fxensemble": FXEnsemble,
}


class Instructor:
    def __init__(self, opt):
        self.opt = opt

        self.transformer_tokenizers = {}
        self.transformer_models = {}

        own_model_class = OWN_MODELNAME2CLASS[opt.own_model_name]

        # create transformer models
        for pretrained_weights_name in own_model_class.get_language_models():
            self.create_transformer_model(pretrained_weights_name)
        logger.info("initialized transformer tokenizers and models")

        # setup own tokenizer
        self.own_tokenizer = FXEasyTokenizer(
            self.transformer_tokenizers,
            self.opt.max_seq_len,
            self.opt.knowledgesources,
            self.opt.is_use_natural_target_phrase_for_spc,
        )

        self.trainset = None
        self.devset = None
        self.testset = None
        self.all_datasets = None

        if self.opt.training_mode:
            self.load_datasets()

        # setup own model
        own_model_object = own_model_class(self.transformer_models, self.opt)
        own_model_object = own_model_object.to(self.opt.device)
        if self.opt.state_dict:
            Download.download(own_model_class)
            logger.info("loading weights from %s...", self.opt.state_dict)
            state_dict = torch.load(
                self.opt.state_dict, map_location=self.opt.device
            )
            own_model_object.load_state_dict(state_dict)
            logger.info("done")
        self.own_model = own_model_object
        logger.info("initialized own model")

        self.evaluator = Evaluator(
            SentimentClasses.get_sorted_expected_label_values(),
            SentimentClasses.get_polarity_associations(),
            self.opt.snem,
        )

        self._print_args()

    def _load_dataset(self, path, coref_mode):
        return FXDataset(
            path,
            self.opt.data_format,
            self.own_tokenizer,
            SentimentClasses.get_polarity_associations(),
            SentimentClasses.get_polarity_associations_inverse(),
            SentimentClasses.get_sorted_expected_label_names(),
            self.opt.single_targets,
            coref_mode,
            self.opt.devmode,
            self.opt.ignore_parsing_errors,
        )

    def load_datasets(self):
        logger.info(
            "loading datasets {} from {}".format(
                self.opt.dataset_name, self.opt.dataset_path
            )
        )
        self.trainset = self._load_dataset(
            self.opt.dataset_path + "train.jsonl",
            coref_mode=self.opt.coref_mode_in_training,
        )
        self.devset = self._load_dataset(
            self.opt.dataset_path + "dev.jsonl", coref_mode="ignore"
        )
        self.testset = self._load_dataset(
            self.opt.dataset_path + "test.jsonl", coref_mode="ignore"
        )
        self.all_datasets = [self.trainset, self.devset, self.testset]
        logger.info("loaded datasets from {}".format(self.opt.dataset_path))

        all_datasets_overflow_counter = Counter()
        for dataset in self.all_datasets:
            all_datasets_overflow_counter += dataset.count_sequence_overflow
        logger.info(
            f"count truncated sequences in total: {all_datasets_overflow_counter}"
        )

    def _print_args(self):
        n_trainable_params, n_nontrainable_params = 0, 0
        for p in self.own_model.parameters():
            n_params = torch.prod(torch.tensor(p.shape))
            if p.requires_grad:
                n_trainable_params += n_params
            else:
                n_nontrainable_params += n_params
        logger.info(
            "n_trainable_params: {0}, n_nontrainable_params: {1}".format(
                n_trainable_params, n_nontrainable_params
            )
        )
        logger.info("> training arguments:")
        for arg in vars(self.opt):
            logger.info(">>> {0}: {1}".format(arg, getattr(self.opt, arg)))

    def create_transformer_model(
        self, pretrained_weights_name: str, only_model: bool = False
    ):
        logger.info(
            "creating model for weights name: {}".format(pretrained_weights_name)
        )

        if not only_model:
            model_info = TRANSFORMER_MODELS_INFO[pretrained_weights_name]
            tokenizer_class = model_info["tokenizer_class"]
            model_class = model_info["model_class"]

            if (
                self.opt.pretrained_model_name
                and self.opt.pretrained_model_name is not None
                and self.opt.pretrained_model_name != "default"
            ):
                model_path = self.opt.pretrained_model_name
            else:
                model_path = pretrained_weights_name
            logger.info("using model_path: %s", model_path)

            self.transformer_tokenizers[
                pretrained_weights_name
            ] = tokenizer_class.from_pretrained(model_path)
            self.transformer_models[
                pretrained_weights_name
            ] = model_class.from_pretrained(model_path, output_hidden_states=True)

    def _reset_params_of_own_model(self):
        for child in self.own_model.children():
            if not issubclass(
                child.__class__, PreTrainedModel
            ):  # if type(child) != BertModel:  # skip bert params
                for p in child.parameters():
                    if p.requires_grad:
                        if len(p.shape) > 1:
                            self.opt.initializer(p)
                        else:
                            stdv = 1.0 / math.sqrt(p.shape[0])
                            torch.nn.init.uniform_(p, a=-stdv, b=stdv)

    def _create_prepare_model_path_to_store_state_dict(self, snem, epoch):
        selected_model_filename = "{0}_{1}_val_{2}_{3}_epoch{4}".format(
            self.opt.own_model_name,
            self.opt.dataset_name,
            self.opt.snem,
            round(snem, 4),
            epoch,
        )

        pathdir = os.path.join(self.opt.experiment_path, "state_dict")

        os.makedirs(pathdir, exist_ok=True)
        selected_model_path = os.path.join(pathdir, selected_model_filename)

        return selected_model_filename, selected_model_path

    def select_inputs(self, sample_batched, is_single_item=False):
        """
        The input to this function is a batch of the dataset, aggregated by pytorch's
        DataLoader. The DataLoader performs complex aggregation of multiple items of the
        batch into one vector, e.g.:
        - a field containing a single number or boolean in each item will become a tensor of size
          k, where k is the number of items in the batch. Examples: polarity
        - a field containing a vector (list of) numbers of some size up to m, will
          become a tensor of size (k, m). Examples: target_mask, text_ids_with_special_tokens
        - a field containing a str, will become a list of str of size k. Examples: label

        The function at hand then selects only those fields that are one of
         [FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS, FIELD_TARGET_MASK], for each LM tokenizer
         defined in the FXEasyTokenizer (this information is setup during initializing
          the instructor), specifically in self.own_tokenizer.tokenizers_name_and_obj.
        """
        inputs = []
        params2index = {}
        index = 0
        for weights_name, field_name in self.own_model.get_input_field_ids():
            seq = sample_batched[weights_name][field_name]
            if is_single_item:
                # if we only have a single item, we need to add a virtual, singleton
                # batch dimension
                seq = seq.unsqueeze(0)
            # move to device and append to list
            inputs.append(seq.to(self.opt.device))
            # store in param dict for easy access later
            param_id = (weights_name, field_name)
            params2index[param_id] = index
            index += 1

        FXDataset.set_params2index(params2index)

        return inputs

    def _train(self, criterion, optimizer, train_data_loader, dev_data_loader):
        global_step = 0
        selected_model_path = None
        selected_model_filename = None
        selected_model_dev_stats = None

        # initialize the early_stopping object
        early_stopping = EarlyStopping()

        # skip training
        if self._is_random_model():
            logger.info(
                "skipping training because random model: %s", self.opt.own_model_name
            )
            # perform one fake intraining evaluation so that we can save the model and
            # use that for testset evaluation
            # set epoch to a bit higher than max number of epochs to ensure that the
            # model is being saved
            fake_epoch = self.opt.num_epoch + 2
            (
                selected_model_path,
                selected_model_filename,
                selected_model_dev_stats,
            ) = self._intraining_evaluation_and_model_save(
                dev_data_loader,
                early_stopping,
                fake_epoch,
                selected_model_path,
                selected_model_filename,
                selected_model_dev_stats,
            )
        else:
            for epoch in range(self.opt.num_epoch):
                logger.info(">" * 100)
                logger.info(
                    "epoch: {} (num_epoch: {})".format(epoch, self.opt.num_epoch)
                )
                num_correct_predictions_total = 0
                num_predictions_total = 0
                num_examples_total = 0
                loss_total = 0

                # switch model to training mode
                self.own_model.train()

                # train on batches
                for i_batch, sample_batched in enumerate(train_data_loader):
                    global_step += 1
                    # clear gradient accumulators
                    optimizer.zero_grad()
                    # select only relevant fields
                    inputs = self.select_inputs(sample_batched)
                    targets = sample_batched["polarity"].to(self.opt.device)

                    # invoke the model
                    outputs = self.own_model(inputs)

                    # some models return multiple objects
                    if self.opt.own_model_name in ["contrasting"]:
                        outputs, cross_weight = outputs

                    # since the following logic is based on multi targets, but in case of
                    # single_targets=True this dimension is removed (since the state of the
                    # art models for single target per item, such as LCF, were not adapted
                    # to multiple targets, and thus output only a two dimensional tensor
                    # (batch, polarity)), we need to add a singleton dimension representing
                    # the target dimension
                    if self.opt.single_targets:
                        outputs = outputs.unsqueeze(1)

                    # permute to match k-dimensional crossentropyloss,
                    # see https://pytorch.org/docs/stable/nn.html#crossentropyloss
                    outputs = outputs.permute(0, 2, 1)

                    # apply loss
                    if self.opt.loss == "sequence":
                        # sequence loss requires targets mask (indev: currently, this will
                        # work only for the "seq2seq" model)
                        text_bert_indices_targets_mask = FXDataset.get_input_by_params(
                            inputs,
                            get_default_lm(),
                            FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK,
                        )
                        loss = criterion(
                            outputs, targets, text_bert_indices_targets_mask
                        )
                    elif self.opt.loss == "crossentropy_crossweight":
                        text_bert_indices_targets_mask = FXDataset.get_input_by_params(
                            inputs,
                            get_default_lm(),
                            FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK,
                        )
                        loss = criterion(
                            outputs,
                            targets,
                            cross_weight,
                            text_bert_indices_targets_mask,
                        )
                    else:
                        loss = criterion(outputs, targets)

                    loss.backward()
                    optimizer.step()

                    # calculate training stats
                    # get predicted classes
                    if self.opt.loss == "sequence":
                        predicted_classes = self._get_classes_from_sequence_output(
                            outputs, text_bert_indices_targets_mask
                        )
                    else:
                        predicted_classes = torch.argmax(outputs, dim=1)

                    # calculate number of correctly predicted classes and sum it. here, we
                    # can ignore SentimentClasses.FILLUP_POLARITY_VALUE because when in
                    # a cell in "targets" the value is
                    # SentimentClasses.FILLUP_POLARITY_VALUE, the corresponding value in
                    # predicted_classes will never be the same (so this will never attribute
                    # to a correct prediction)
                    num_correct_predictions = (
                        (predicted_classes == targets).sum().item()
                    )
                    # to count number of all predictions, we need to ignore those that have
                    # SentimentClasses.FILLUP_POLARITY_VALUE in a cell of "targets"
                    # note, that we do not use the number of examples but the number of
                    # predictions, since one training example can contain one or more
                    # targets
                    num_predictions = targets != SentimentClasses.FILLUP_POLARITY_VALUE
                    num_predictions = num_predictions.sum().item()
                    # ensure the number of predictions in this batch is equal to the number
                    # of targets defined by the dataset.py
                    assert num_predictions == sample_batched["num_targets"].sum().item()
                    # get number of examples (we can use more or less any field in
                    # sample_batched to get the number of examples)
                    num_examples = len(sample_batched["num_targets"])

                    # update stats
                    num_correct_predictions_total += num_correct_predictions
                    num_predictions_total += num_predictions
                    loss_total += loss.item() * num_predictions
                    num_examples_total += num_examples
                    if global_step % self.opt.log_step == 0:
                        # update accuracy
                        train_acc = (
                            num_correct_predictions_total / num_predictions_total
                        )
                        train_loss = loss_total / num_predictions_total
                        logger.info(
                            "loss: {:.4f}, acc: {:.4f}, ex.: {}, pred.: {}/{}".format(
                                train_loss,
                                train_acc,
                                num_examples_total,
                                num_correct_predictions_total,
                                num_predictions_total,
                            )
                        )

                # completed all batches of current epoch. now, perform a devset
                # evaluation and if best model so far, save it
                (
                    selected_model_path,
                    selected_model_filename,
                    selected_model_dev_stats,
                ) = self._intraining_evaluation_and_model_save(
                    dev_data_loader,
                    early_stopping,
                    epoch,
                    selected_model_path,
                    selected_model_filename,
                    selected_model_dev_stats,
                )

                if early_stopping.early_stop and self.opt.use_early_stopping:
                    logger.info(
                        "early stopping after {} epochs without improvement, total epochs: {} of {}".format(
                            early_stopping.patience, epoch, self.opt.num_epoch
                        )
                    )
                    break

        return selected_model_path, selected_model_filename, selected_model_dev_stats

    def _intraining_evaluation_and_model_save(
        self,
        dev_data_loader,
        early_stopping,
        epoch,
        selected_model_path,
        selected_model_filename,
        selected_model_dev_stats,
    ):
        dev_stats = self._evaluate(dev_data_loader)
        dev_snem = dev_stats[self.opt.snem]
        self.evaluator.print_stats(dev_stats, "dev during training")
        early_stopping(dev_snem)

        (
            has_stored,
            _selected_model_filename,
            _selected_model_path,
            _selected_model_dev_stats,
        ) = self._save_model_state_dict(early_stopping, epoch, dev_stats)
        if has_stored:
            selected_model_dev_stats = _selected_model_dev_stats
            selected_model_path = _selected_model_path
            selected_model_filename = _selected_model_filename

        return selected_model_path, selected_model_filename, selected_model_dev_stats

    def _save_model_state_dict(self, early_stopping, epoch: int, dev_stats):
        dev_snem = dev_stats[self.opt.snem]

        has_stored = False
        selected_model_filename = None
        selected_model_path = None
        selected_model_dev_stats = None

        if self.opt.eval_only_after_last_epoch:
            if epoch >= self.opt.num_epoch - 1:
                # return the model that was trained through all epochs as the selected model
                logger.info("all epochs finished, saving model to disk...")

                has_stored = True
                selected_model_dev_stats = dev_stats

                (
                    selected_model_filename,
                    selected_model_path,
                ) = self._create_prepare_model_path_to_store_state_dict(dev_snem, epoch)
                torch.save(self.own_model.state_dict(), selected_model_path)
                logger.info(">> saved: {}".format(selected_model_path))

                # save confusion matrices
                filepath_stats_base = os.path.join(
                    self.opt.experiment_path, "statistics", selected_model_filename
                )
                if not filepath_stats_base.endswith("/"):
                    filepath_stats_base += "/"
                os.makedirs(filepath_stats_base, exist_ok=True)
                create_save_plotted_confusion_matrix(
                    dev_stats["confusion_matrix"],
                    expected_labels=SentimentClasses.get_sorted_expected_label_values(),
                    basepath=filepath_stats_base,
                )
                logger.debug(
                    "created confusion matrices in path: {}".format(filepath_stats_base)
                )
        else:
            # return the best model during any epoch
            if early_stopping.flag_has_score_increased_since_last_check:
                logger.info("model yields best performance so far, saving to disk...")
                has_stored = True
                selected_model_dev_stats = dev_stats
                (
                    selected_model_filename,
                    selected_model_path,
                ) = self._create_prepare_model_path_to_store_state_dict(dev_snem, epoch)

                torch.save(self.own_model.state_dict(), selected_model_path)
                logger.info(">> saved: {}".format(selected_model_path))

                # save confusion matrices
                filepath_stats_base = os.path.join(
                    self.opt.experiment_path, "statistics", selected_model_filename
                )
                if not filepath_stats_base.endswith("/"):
                    filepath_stats_base += "/"
                os.makedirs(filepath_stats_base, exist_ok=True)
                create_save_plotted_confusion_matrix(
                    dev_stats["confusion_matrix"],
                    expected_labels=SentimentClasses.get_sorted_expected_label_values(),
                    basepath=filepath_stats_base,
                )
                logger.debug(
                    "created confusion matrices in path: {}".format(filepath_stats_base)
                )
        return (
            has_stored,
            selected_model_filename,
            selected_model_path,
            selected_model_dev_stats,
        )

    def _get_classes_from_sequence_output(
        self, model_outputs, text_bert_indices_targets_mask
    ):
        """

        :param model_outputs: Tensor of shape:
            batch, numclasses, seqlen
        :param text_bert_indices_targets_mask: Tensor of shape:
            batch, targets, seqlen
        :return: Tensor of shape:
            batch, targets
        """
        num_targets = text_bert_indices_targets_mask.shape[1]
        num_classes = model_outputs.shape[1]

        # prepare model output
        model_outputs = model_outputs.unsqueeze(1)
        model_outputs = model_outputs.repeat(1, num_targets, 1, 1)
        # new shape: batch, targets, numclasses, seqlen

        # prepare text_bert_indices_targets_mask
        text_bert_indices_targets_mask = text_bert_indices_targets_mask.unsqueeze(2)
        text_bert_indices_targets_mask = text_bert_indices_targets_mask.repeat(
            1, 1, num_classes, 1
        )
        # new shape: batch, targets, numclasses, seqlen

        # multiply so that we will only have the values model output values that belong
        # to a target node (1's for output nodes with a target in text_bert_indices...
        # 0's for non-targets)
        only_target_outputs = model_outputs * text_bert_indices_targets_mask
        # shape: batch, targets, numclasses, seqlen

        # aggregate, e.g., mean
        # note that in principle any way of reducing (or aggregating) could be used here
        # e.g., also sum or max. however, the reducing method should as good as possible
        # match the way of calculating the sequence based loss (seq2seqloss.py). our
        # seq-based loss currently equally assigns the loss to each target output node
        # so that in my opinion the unweighted mean is most appropriate here for
        # reduction, since it also equally uses the output of each target node
        only_target_outputs_aggregated = only_target_outputs.mean(3)
        # shape: batch, targets, classes

        predicted_classes = torch.argmax(only_target_outputs_aggregated, dim=2)
        # predicted_classes shape: batch, targets
        return predicted_classes

    def _evaluate(self, data_loader, get_examples=False, basepath=None):
        t_labels_all = None
        t_outputs_all = None
        t_text_bert_indices_targets_mask_all = None
        t_texts_all = []

        # switch model to evaluation mode
        self.own_model.eval()

        with torch.no_grad():
            for t_batch, t_sample_batched in enumerate(data_loader):
                t_inputs = self.select_inputs(t_sample_batched)
                t_labels = t_sample_batched["polarity"].to(self.opt.device)
                t_text_bert_indices_targets_mask = FXDataset.get_input_by_params(
                    t_inputs,
                    self.own_model.get_language_models()[0],
                    FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK,
                )
                assert len(self.own_model.get_language_models()) == 1, (
                    f"multiple language models per own model are currently not supported: "
                    f"{self.own_model.get_language_models()}, "
                    f"{len(self.own_model.get_language_models())}"
                )

                t_texts = t_sample_batched["text"]
                # t_targets = t_sample_batched["orig_targets"]

                # invoke the model
                t_outputs = self.own_model(t_inputs)
                # some models return multiple objects
                if self.opt.own_model_name in ["contrasting"]:
                    t_outputs, t_cross_weight = t_outputs

                if self.opt.single_targets:
                    t_outputs = t_outputs.unsqueeze(1)

                if t_labels_all is None:
                    t_labels_all = t_labels
                    t_outputs_all = t_outputs
                    t_text_bert_indices_targets_mask_all = (
                        t_text_bert_indices_targets_mask
                    )
                else:
                    t_labels_all = torch.cat((t_labels_all, t_labels), dim=0)
                    t_outputs_all = torch.cat((t_outputs_all, t_outputs), dim=0)
                    t_text_bert_indices_targets_mask_all = torch.cat(
                        (
                            t_text_bert_indices_targets_mask_all,
                            t_text_bert_indices_targets_mask,
                        ),
                        dim=0,
                    )

                t_texts_all.extend(t_texts)
                # t_targets_all.extend(t_targets)

        y_true = t_labels_all.cpu()

        if self.opt.loss == "sequence":
            # we first need to convert sequence-based predictions to target-based
            # predictions
            # permute to match requirement of _get_classes_from_sequence_output
            t_outputs_all = t_outputs_all.permute(0, 2, 1)
            y_pred = self._get_classes_from_sequence_output(
                t_outputs_all, t_text_bert_indices_targets_mask_all
            ).cpu()
        else:
            # softmax: get predictions from outputs
            # have to take the 3rd (dim=2) dimension
            y_pred = torch.argmax(t_outputs_all, dim=2).cpu()

        stats = self.evaluator.calc_statistics(y_true, y_pred)

        if get_examples:
            self.evaluator.write_error_table(
                y_true, y_pred, t_texts_all, basepath + "errortable.jsonl",
            )

        return stats

    def get_normalized_inv_class_frequencies(self):
        inv_freqs = []
        for label_name in self.sorted_expected_label_names:
            inv_freq_of_class = 1.0 / self.testset.label_counter[label_name]
            inv_freqs.append(inv_freq_of_class)

        sum_of_inv_freqs = sum(inv_freqs)
        for i in range(len(inv_freqs)):
            inv_freqs[i] = inv_freqs[i] / sum_of_inv_freqs

        return inv_freqs

    def _is_random_model(self):
        return self.opt.own_model_name in ["random_multi", "random_single"]

    def run(self):
        # balancing modes
        class_weights = None
        sampler_train = None
        assert self.opt.balancing is None, (
            "balancing is currently only implemented for None. if we want to have it, "
            "we need to come up with a logic that considers that an item can have "
            "multiple polarities in mtsc-mode"
        )
        if self.opt.balancing == "lossweighting":
            inv_class_freqs = self.get_normalized_inv_class_frequencies()
            logger.info("weighting losses of classes: {}".format(inv_class_freqs))
            class_weights = torch.tensor(inv_class_freqs).to(self.opt.device)
        elif self.opt.balancing == "oversampling":
            sampler_train = RandomOversampler(self.trainset, self.opt.seed)

        if self.opt.loss == "crossentropy":
            criterion = nn.CrossEntropyLoss(
                weight=class_weights,
                ignore_index=SentimentClasses.FILLUP_POLARITY_VALUE,
            )
        elif self.opt.loss == "crossentropy_lsr":
            criterion = CrossEntropyLoss_LSR(
                self.opt.device, smoothing_value=0.2, weight=class_weights
            )
        elif self.opt.loss == "crossentropy_crossweight":
            criterion = CrossEntropyLoss_CrossWeight(
                device=self.opt.device,
                weight=class_weights,
                ignore_index=SentimentClasses.FILLUP_POLARITY_VALUE,
            )
        elif self.opt.loss == "sequence":
            criterion = SequenceLoss(device=self.opt.device, weight=class_weights)
        else:
            raise ValueError("loss unknown: {}".format(self.opt.loss))

        # optimizer
        if self._is_random_model():
            logger.info("initialize optimizer")
            optimizer = None
        else:
            logger.error("check if the correct parameters are retrieved here!")
            _params = filter(lambda p: p.requires_grad, self.own_model.parameters())
            optimizer = self.opt.optimizer(
                _params, lr=self.opt.learning_rate, weight_decay=self.opt.l2reg
            )

        # data loaders
        if sampler_train:
            train_data_loader = DataLoader(
                dataset=self.trainset,
                batch_size=self.opt.batch_size,
                sampler=sampler_train,
            )
        else:
            train_data_loader = DataLoader(
                dataset=self.trainset, batch_size=self.opt.batch_size, shuffle=True
            )
        dev_data_loader = DataLoader(
            dataset=self.devset, batch_size=self.opt.batch_size, shuffle=False
        )
        test_data_loader = DataLoader(
            dataset=self.testset, batch_size=self.opt.batch_size, shuffle=False
        )

        # start training
        self._reset_params_of_own_model()
        logger.info("starting training...")
        time_training_start = time.time()
        best_model_path, best_model_filename, selected_model_dev_stats = self._train(
            criterion, optimizer, train_data_loader, dev_data_loader
        )
        time_training_elapsed_mins = (time.time() - time_training_start) // 60
        logger.info(
            "training finished. duration [mins]: {}".format(time_training_elapsed_mins)
        )

        self.perform_post_training_actions(
            best_model_path,
            best_model_filename,
            test_data_loader,
            selected_model_dev_stats,
            time_training_elapsed_mins,
        )

    def get_serializable_stats(self, stats):
        sstats = stats.copy()
        sstats["recalls_of_classes"] = stats["recalls_of_classes"].tolist()
        sstats["confusion_matrix"] = stats["confusion_matrix"].tolist()
        return sstats

    def get_serializable_opts(self):
        opts = vars(self.opt)
        sopts = opts.copy()
        del sopts["optimizer"]
        del sopts["initializer"]
        del sopts["device"]
        return sopts

    def perform_post_training_actions(
        self,
        selected_model_path,
        selected_model_filename,
        test_data_loader,
        selected_model_dev_stats,
        time_training_elapsed_mins,
    ):
        logger.info(
            "loading selected model from training: {}".format(selected_model_path)
        )
        self.own_model.load_state_dict(torch.load(selected_model_path))

        logger.info("evaluating selected model on test-set")
        # set model into evaluation mode (cf. https://pytorch.org/docs/stable/nn.html#torch.nn.Module.train)
        self.own_model.eval()

        # do the actual evaluation
        filepath_stats_prefix = os.path.join(
            self.opt.experiment_path, "statistics", selected_model_filename
        )
        os.makedirs(filepath_stats_prefix, exist_ok=True)
        if not filepath_stats_prefix.endswith("/"):
            filepath_stats_prefix += "/"

        test_stats = self._evaluate(
            test_data_loader, get_examples=True, basepath=filepath_stats_prefix
        )
        test_snem = test_stats[self.opt.snem]

        self.evaluator.print_stats(test_stats, "evaluation on test-set")

        # save dev and test results
        experiment_results = {}
        experiment_results["test_stats"] = self.get_serializable_stats(test_stats)
        experiment_results["dev_stats"] = self.get_serializable_stats(
            selected_model_dev_stats
        )
        experiment_results["options"] = self.get_serializable_opts()
        experiment_results["time_training_elapsed_mins"] = time_training_elapsed_mins

        experiment_results_path = os.path.join(
            self.opt.experiment_path, "experiment_results.jsonl"
        )
        with jsonlines.open(experiment_results_path, "w") as writer:
            writer.write(experiment_results)

        # save confusion matrices
        test_confusion_matrix = test_stats["confusion_matrix"]

        create_save_plotted_confusion_matrix(
            test_confusion_matrix,
            expected_labels=SentimentClasses.get_sorted_expected_label_values(),
            basepath=filepath_stats_prefix,
        )

        logger.info("finished execution of this run. exiting.")

        # print snem pad_value to stdout, for the controller to parse it
        print(test_snem)


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean pad_value expected.")


def _setup_cuda(opt):
    logger.info("cuda information")
    logger.info("scc SGE_GPU: %s", os.environ.get("SGE_GPU"))
    logger.info("arg: cuda device: %s", opt.device)

    if torch.cuda.is_available():
        logger.info("cuda is available on this machine")
        if opt.device == "cpu":
            logger.info("forced use of cpu")
            opt.device = torch.device("cpu")
        elif opt.device is None:
            logger.info("no cuda device defined by argument, using default cuda:0")
            opt.device = torch.device("cuda:0")
        else:
            assert opt.device and opt.device != "cpu"
            logger.info("using cuda device defined by argument")
            opt.device = torch.device(opt.device)
    else:
        opt.device = torch.device("cpu")

    if opt.device.type == "cuda":
        logger.info("GPGPU enabled. CUDA dev index: %s", opt.device.index)
        logger.info(
            "using GPU (cuda memory allocated: %s)", torch.cuda.memory_allocated()
        )
    else:
        logger.info("using CPU")


def _die_gracefully(reason: str):
    logger.info("grateful exit:")
    logger.info(reason)
    # traceback.print_stack()
    sys.exit(99)


def check_arguments(opt):
    """
    Perform simple check of arguments passed to the tool. Idea is to die gratefully as
    early as possible with a special RC = -99 if conditions are not met, e.g., neither
    single_targets nor multi_targets mode is enabled, or if single_targets mode is used
    with a loss function for multi targets, etc.

    :param opt:
    :return:
    """

    if not (opt.single_targets != opt.multi_targets):
        _die_gracefully("you must use either " "single_targets or multi_targets")

    if opt.eval_only_after_last_epoch:
        if opt.use_early_stopping:
            _die_gracefully(
                "early stopping and eval only after last epoch both enabled"
            )

    # loss
    # some easy validation
    if opt.own_model_name in ["seq2seq", "seq2seq_withouttargetmask"]:
        if opt.loss != "sequence":
            _die_gracefully(f"model_name: {opt.own_model_name}, loss: {opt.loss}")
    else:
        if opt.loss == "sequence":
            _die_gracefully(f"model_name: {opt.own_model_name}, loss: {opt.loss}")

    if opt.own_model_name in ["contrasting"]:
        if opt.loss != "crossentropy_crossweight":
            _die_gracefully(f"model_name: {opt.own_model_name}, loss: {opt.loss}")
    else:
        if opt.loss == "crossentropy_crossweight":
            _die_gracefully(f"model_name: {opt.own_model_name}, loss: {opt.loss}")

    # check knowledge source
    assert type(opt.knowledgesources) == tuple, "opt.knowledgesources is not tuple"
    for source in opt.knowledgesources:
        if source not in [
            "bingliu_opinion",
            "mpqa_subjectivity",
            "nrc_emotions",
            "liwc",
            "zeros",
        ]:
            _die_gracefully(f"incorrect source: {source} (of '{opt.knowledgesources}')")

    # coref mode
    # single-target mode: one target per example
    # coref ignore: as before
    #       in_targetmask: just add to each target's targetmask (more 1's in there)
    #       additional_ex: create additional examples
    # multi-target mode: k targets per example
    # coref ignore: as before
    #       in_targetmask: see above
    #       additional_ex: create additional examples
    # but all of this ONLY for train set!
    assert opt.coref_mode_in_training in [
        "ignore",
        "in_targetmask",
        "additional_examples",
    ]


def post_process_arguments(opt):
    # if neither single nor multi targets is enabled, assume default, i.e., single
    if not opt.single_targets and not opt.multi_targets:
        opt.single_targets = True

    # post process
    if type(opt.knowledgesources) == str:
        # should be a list but sometimes this does not work...
        opt.knowledgesources = opt.knowledgesources.split(" ")
    elif isinstance(opt.knowledgesources, Iterable) and len(opt.knowledgesources) == 1:
        # actually, if multiple knowledge sources are passed by controller, they will be
        # treated as a quoted str so that a list of 1 item is created
        opt.knowledgesources = opt.knowledgesources[0].split(" ")
    opt.knowledgesources = tuple(opt.knowledgesources)

    for eks in opt.knowledgesources:
        if eks == "nrc_emotions":
            num_categories = get_num_nrc_emotions()
        elif eks == "mpqa_subjectivity":
            num_categories = get_num_mpqa_subjectivity_polarities()
        elif eks == "bingliu_opinion":
            num_categories = get_num_bingliu_polarities()
        elif eks == "liwc":
            from knowledge.liwc.liwc import get_num_liwc_categories

            num_categories = get_num_liwc_categories()
        elif eks == "zeros":
            num_categories = get_num_zero_dimensions()
        else:
            raise NotImplementedError(f"unknown knowledgesource: {eks}")
        FXEasyTokenizer.NUM_CATEGORIES_OF_SELECTED_KNOWLEDGE_SOURCES += num_categories
        logger.info(
            "updated total number of categories to %s with EKS %s",
            FXEasyTokenizer.NUM_CATEGORIES_OF_SELECTED_KNOWLEDGE_SOURCES,
            eks,
        )


def prepare_and_start_instructor(opt):
    # overwrite default lm
    set_default_lm(opt.default_lm)
    logger.info("set default language model to %s", get_default_lm())

    post_process_arguments(opt)

    # check arguments
    check_arguments(opt)

    if opt.targetclasses == "newsmtsc3":
        SentimentClasses.Sentiment3ForNewsMtsc()
    elif opt.targetclasses == "newsmtsc3strong":
        SentimentClasses.SentimentStrong3ForNewsMtsc()
    elif opt.targetclasses == "newsmtsc3weak":
        SentimentClasses.SentimentWeak3ForNewsMtsc()
    else:
        raise ValueError("")
    assert (
        opt.polarities_dim == -1
    ), "polarities_dim needs to be -1 for automatic detection"
    opt.polarities_dim = len(SentimentClasses.SENTIMENT_CLASSES)
    logger.info("set number of polarity classes to %s", opt.polarities_dim)

    if not opt.balancing or opt.balancing == "None":
        opt.balancing = None

    if opt.seed is None:
        opt.seed = int(time.time())
        logger.info("no random seed was given, using system time")
    logger.info("setting random seed: {}".format(opt.seed))
    random.seed(opt.seed)
    numpy.random.seed(opt.seed)
    torch.manual_seed(opt.seed)
    torch.cuda.manual_seed(opt.seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False

    initializers = {
        "xavier_uniform_": torch.nn.init.xavier_uniform_,
        "xavier_normal_": torch.nn.init.xavier_normal,
        "orthogonal_": torch.nn.init.orthogonal_,
    }
    optimizers = {
        "adadelta": torch.optim.Adadelta,  # default lr=1.0
        "adagrad": torch.optim.Adagrad,  # default lr=0.01
        "adam": torch.optim.Adam,  # default lr=0.001
        "adamax": torch.optim.Adamax,  # default lr=0.002
        "asgd": torch.optim.ASGD,  # default lr=0.01
        "rmsprop": torch.optim.RMSprop,  # default lr=0.01
        "sgd": torch.optim.SGD,
    }

    # support for custom pretrained language models, e.g., domain adapted models
    if not opt.pretrained_model_name or opt.pretrained_model_name == "default":
        opt.pretrained_model_name = None
    else:
        opt.pretrained_model_name = os.path.join(
            opt.base_path, "pretrained_models", opt.pretrained_model_name
        )

    # statedicts
    if opt.state_dict is not None:
        if os.path.isfile(opt.state_dict) or opt.state_dict == "pretrained":
            # if exists or using pretrained, nothing to do
            pass
        else:
            # if not exists, prefix default path
            new_path = os.path.join("pretrained_models", "state_dicts", opt.state_dict)
            opt.state_dict = new_path
    else:
        # if not defined, nothing to do
        pass

    if not opt.experiment_path:
        opt.experiment_path = "."
    if not opt.experiment_path.endswith("/"):
        opt.experiment_path = opt.experiment_path + "/"

    _setup_cuda(opt)

    if opt.training_mode:
        logger.debug("dataset_path not defined, creating from dataset_name...")
        opt.dataset_path = os.path.join("datasets", opt.dataset_name)
        if not opt.dataset_path.endswith("/"):
            opt.dataset_path = opt.dataset_path + "/"
        logger.debug(
            "dataset_path created from dataset_name: {}".format(opt.dataset_path)
        )

        # set dataset_path to include experiment_path
        opt.dataset_path = os.path.join(opt.experiment_path, opt.dataset_path)

    ins = Instructor(opt)

    if opt.training_mode:
        opt.initializer = initializers[opt.initializer]
        opt.optimizer = optimizers[opt.optimizer]

        ins.run()
    else:
        return ins


def parse_arguments(override_args=False, overwrite_logging_level=None):
    """
    Where applicable, for each argument a default is defined that led to the best
    results in our experiments for grutsc (which itself is set as a default for the
    corresponding argument).
    :param override_args:
    :param overwrite_logging_level:
    :return:
    """

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--training_mode", type=str2bool, nargs="?", const=True, default=True
    )
    parser.add_argument("--own_model_name", type=str)
    parser.add_argument(
        "--dataset_name",
        default=None,
        type=str,
        help="name of the sub-folder in 'datasets' containing files named [train,dev,test].jsonl",
    )
    parser.add_argument("--data_format", default="newsmtsc", type=str)
    parser.add_argument("--optimizer", default="adam", type=str)
    parser.add_argument("--initializer", default="xavier_uniform_", type=str)
    parser.add_argument(
        "--learning_rate",
        default=2e-5,
        type=float,
        help="try 5e-5, 2e-5 for BERT, 1e-3 for others",
    )
    parser.add_argument("--dropout", default=0.1, type=float)
    parser.add_argument("--l2reg", default=0.01, type=float)
    parser.add_argument(
        "--num_epoch",
        default=3,
        type=int,
        help="try larger number for non-BERT models",
    )
    parser.add_argument(
        "--batch_size", default=16, type=int, help="try 16, 32, 64 for BERT models"
    )
    parser.add_argument("--log_step", default=5, type=int)
    parser.add_argument("--max_seq_len", default=150, type=int)
    parser.add_argument("--polarities_dim", default=-1, type=int)
    parser.add_argument(
        "--device",
        default=None,
        type=str,
        help="e.g., cuda:0; if None, CPU will be used",
    )
    parser.add_argument(
        "--seed", default=None, type=int, help="set seed for reproducibility"
    )
    parser.add_argument(
        "--local_context_focus",
        default="cdm",
        type=str,
        help="local context focus mode, cdw or cdm",
    )
    parser.add_argument(
        "--SRD",
        default=3,
        type=int,
        help="semantic-relative-distance, see the paper of LCF-BERT " "model",
    )
    parser.add_argument(
        "--snem", default="f1_macro", help="see evaluator.py for valid options"
    )
    parser.add_argument(
        "--devmode",
        type=str2bool,
        nargs="?",
        const=True,
        default=False,
        help="devmode, default off, enable by using True",
    )
    parser.add_argument(
        "--experiment_path",
        default="./experiments/default",
        type=str,
        help="if defined, all data will be read from / saved to a folder in the experiments folder",
    )
    parser.add_argument("--balancing", type=str, default=None)
    parser.add_argument("--spc_lm_representation", type=str, default="mean_last")
    parser.add_argument(
        "--spc_input_order",
        type=str,
        default="text_target",
        help="SPC: order of input; target_text " "or text_target",
    )
    parser.add_argument(
        "--use_early_stopping", type=str2bool, nargs="?", const=True, default=False
    )
    parser.add_argument(
        "--eval_only_after_last_epoch",
        type=str2bool,
        nargs="?",
        const=True,
        default=False,
        help="if False, evaluate the best model that was seen during any training epoch. if True, "
        "evaluate only the model that was trained through all num_epoch epochs.",
    )
    parser.add_argument(
        "--pretrained_model_name",
        type=str,
        default="default",
        help="has to be placed in folder pretrained_models",
    )
    parser.add_argument("--state_dict", type=str, default=None)
    parser.add_argument(
        "--single_targets", type=str2bool, nargs="?", const=True, default=False
    )
    parser.add_argument(
        "--multi_targets", type=str2bool, nargs="?", const=True, default=False
    )
    parser.add_argument("--loss", type=str, default="crossentropy")
    parser.add_argument("--targetclasses", type=str, default="newsmtsc3")
    parser.add_argument(
        "--knowledgesources",
        nargs="+",
        default=["nrc_emotions", "mpqa_subjectivity", "bingliu_opinion"],
    )
    parser.add_argument(
        "--is_use_natural_target_phrase_for_spc",
        type=str2bool,
        nargs="?",
        const=True,
        default=False,
    )
    parser.add_argument("--default_lm", type=str, default=ROBERTA_BASE)
    parser.add_argument("--run_id", type=int, default=0)
    parser.add_argument("--coref_mode_in_training", type=str, default="ignore")
    parser.add_argument(
        "--logging", type=str, default="INFO",
    )
    parser.add_argument(
        "--ignore_parsing_errors", type=str2bool, nargs="?", const=True, default=False
    )

    # if own_args == None -> parse_args will use sys.argv
    # if own_args == [] -> parse_args will use this empty list instead
    own_args = None
    if override_args:
        own_args = []

    # create arguments
    opt = parser.parse_args(args=own_args)

    # add basepath
    dir_path = os.path.dirname(os.path.realpath(__file__))
    opt.base_path = dir_path

    # set logging
    if overwrite_logging_level:
        logger.setLevel(overwrite_logging_level)

    return opt


if __name__ == "__main__":
    logger.info("python bin: %s", sys.executable)
    opt = parse_arguments()
    prepare_and_start_instructor(opt)



from .dic import read_dic
from .trie import build_trie, search_trie

try:
    import pkg_resources

    __version__ = pkg_resources.get_distribution("liwc").version
except Exception:
    __version__ = None


def load_token_parser(filepath="knowledge/liwc/data/LIWC2015_English.dic"):
    """
    Reads a LIWC lexicon from a file in the .dic format, returning a tuple of
    (parse, category_names), where:
    * `parse` is a function from a token to a list of strings (potentially
      empty) of matching categories
    * `category_names` is a list of strings representing all LIWC categories in
      the lexicon
    """
    lexicon, category_names = read_dic(filepath)
    trie = build_trie(lexicon)

    def parse_token(token):
        return search_trie(trie, token)

    return parse_token, category_names



import os

import matplotlib.pyplot as plt
import numpy as np
from sklearn.metrics import confusion_matrix

from NewsSentiment.fxlogger import get_logger

logger = get_logger()


def create_save_plotted_confusion_matrix(conf_matrix, expected_labels, basepath):
    ax, title = plot_confusion_matrix(conf_matrix, expected_labels, normalize=False)
    filepath = os.path.join(basepath, 'stats.png')
    plt.savefig(filepath, bbox_inches='tight')
    logger.debug("created confusion matrices in path: {}".format(filepath))


def plot_confusion_matrix(cm, classes, normalize=False, title=None, cmap=plt.cm.Blues):
    """
    This function prints and plots the confusion matrix. Normalization can be applied by setting `normalize=True`.
    based on https://scikit-learn.org/stable/auto_examples/model_selection/plot_confusion_matrix.html#sphx-glr-auto-examples-model-selection-plot-confusion-matrix-py
    """
    if not title:
        if normalize:
            title = 'Normalized confusion matrix'
        else:
            title = 'Confusion matrix, without normalization'

    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
        logger.debug("Normalized confusion matrix")
    else:
        logger.debug('Confusion matrix, without normalization')

    logger.debug(cm)

    fig, ax = plt.subplots()
    im = ax.imshow(cm, interpolation='nearest', cmap=cmap)
    ax.figure.colorbar(im, ax=ax)
    # We want to show all ticks...
    ax.set(xticks=np.arange(cm.shape[1]),
           yticks=np.arange(cm.shape[0]),
           # ... and label them with the respective list entries
           xticklabels=classes, yticklabels=classes,
           title=title,
           ylabel='True label',
           xlabel='Predicted label')

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
             rotation_mode="anchor")

    # Loop over data dimensions and create text annotations.
    fmt = '.2f' if normalize else 'd'
    thresh = cm.max() / 2.
    for i in range(cm.shape[0]):
        for j in range(cm.shape[1]):
            ax.text(j, i, format(cm[i, j], fmt),
                    ha="center", va="center",
                    color="white" if cm[i, j] > thresh else "black")
    fig.tight_layout()

    return ax, title


if __name__ == '__main__':
    y_true = ["cat", "ant", "cat", "cat", "ant", "bird"]
    y_pred = ["ant", "ant", "cat", "cat", "ant", "cat"]
    confmat = confusion_matrix(y_true, y_pred, labels=["ant", "bird", "cat"])

    create_save_plotted_confusion_matrix(confmat, ["ant", "bird", "cat"], '.')



from argparse import Namespace
from typing import Dict

import torch
import torch.nn as nn

from NewsSentiment.consts import *
from NewsSentiment.dataset import FXDataset
from NewsSentiment.models.FXBaseModel import FXBaseModel


class Contrasting(FXBaseModel):
    """
    This model returns a sequence that only contains the hidden states of those output
    nodes that represent a word piece of the target phrase. All other hidden states are
    set to 0. The length of the output sequence is opt.max_seq_len

    From a conceptual perspective, TD-BERT is in some aspects similar (e.g., only the
    output of target-phrase-related nodes is used), but in other aspects not similar or
    it is not quite clear how the authors of TD-BERT implemented them. An email I sent
    to them was not answered yet.
    """

    @staticmethod
    def get_language_models():
        return (BERT_BASE_UNCASED,)

    @staticmethod
    def get_input_field_ids():
        return [
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS),
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK),
        ]

    def __init__(self, transformer_models: Dict, opt: Namespace):
        super(Contrasting, self).__init__()
        self.language_model = transformer_models[BERT_BASE_UNCASED]
        self.dropout = nn.Dropout(opt.dropout)

        self.contrasting_weight_dense = nn.Linear(opt.max_seq_len, opt.max_seq_len,)
        self.dense = nn.Linear(
            self.language_model.config.hidden_size, opt.polarities_dim
        )

    def forward(self, inputs):
        # get inputs
        text_bert_indices = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS
        )
        text_bert_indices_targets_mask = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK
        )
        assert text_bert_indices.shape[1] == 2, "contrasting requires two targets"

        # target a
        a_text_bert_indices = text_bert_indices[:, 0, :]
        a_text_bert_indices_targets_mask = text_bert_indices_targets_mask[:, 0, :]

        # target b
        b_text_bert_indices = text_bert_indices[:, 1, :]
        b_text_bert_indices_targets_mask = text_bert_indices_targets_mask[:, 1, :]

        # bert
        (
            a_last_hidden_states,
            a_pooler_output,
            a_all_hidden_states,
        ) = self.language_model(input_ids=a_text_bert_indices)
        (
            b_last_hidden_states,
            b_pooler_output,
            b_all_hidden_states,
        ) = self.language_model(input_ids=b_text_bert_indices)
        stacked_bert_outs_ab = torch.stack(
            (a_last_hidden_states, b_last_hidden_states), dim=1
        )
        # stacked_bert_outs_ab
        stacked_bert_outs_ab = self.dropout(stacked_bert_outs_ab)
        # shape: batch, 2, seqlen, bertdim

        # create weight
        cross_weight = self.contrasting_weight_dense(text_bert_indices_targets_mask)
        cross_weight = cross_weight.unsqueeze(3).repeat(
            1, 1, 1, stacked_bert_outs_ab.shape[3]
        )
        cross_weight = self.dropout(cross_weight)
        # shape: batch, 2, seqlen

        weighted_stacked_bert_outs_ab = stacked_bert_outs_ab * cross_weight

        # sum
        weighted_stacked_bert_outs_ab = weighted_stacked_bert_outs_ab.sum(dim=2)

        # dense
        logits = self.dense(weighted_stacked_bert_outs_ab)

        return logits, cross_weight



# adapted from ABSA-PyTorch
import torch
import torch.nn as nn

# CrossEntropyLoss for Label Smoothing Regularization
from NewsSentiment.SentimentClasses import SentimentClasses


class CrossEntropyLoss_LSR(nn.Module):
    def __init__(self, device, smoothing_value=0.2, weight=None):
        super(CrossEntropyLoss_LSR, self).__init__()
        self.smoothing_value = smoothing_value
        self.device = device
        self.logSoftmax = nn.LogSoftmax(dim=1)
        self.weight = weight

    def _create_smooth_one_hot_for_true_targets(
        self, labels, batch_size, num_classes, num_targets
    ):
        """
        Produces a smooth one hot encoded tensor where all "false" scalars are set to
        base_prob, and all "true" scalars set to base_prob + 1 - smoothing_value.
        Effectively, for example for three classes, this results in:
        False classes: 0.03 (0.3333*0.1=baseprob)
        True classes: 0.03 + 1 - 0.1 = 0.93
        :param labels: the true classes
        :param batch_size:
        :param num_classes:
        :return:
        """
        # prior label distribution is set uniform (see Sec 3.5,
        # https://arxiv.org/pdf/1902.09314.pdf )
        prior_distribution = 1.0 / float(num_classes)
        # calculate base probability
        base_prob = self.smoothing_value * prior_distribution
        # initialize 0 tensor
        one_hot_label = torch.zeros(
            batch_size, num_classes, num_targets, device=self.device
        )
        # set probability of all classes in all batches and targets to the base prob
        # (in normal one hot encoding, this would be 0 instead)
        one_hot_label = one_hot_label + base_prob

        if self.weight is not None:
            raise NotImplementedError(
                "test this first!!! currently untested with multi tsc"
            )
            one_hot_label = one_hot_label * self.weight

        # iterate over each single item of the batch
        for batch_index in range(batch_size):
            for target_index in range(num_targets):
                # get the class index
                class_index = labels[batch_index, target_index].item()
                if class_index == SentimentClasses.FILLUP_POLARITY_VALUE:
                    # cant set the one hot encoded here, since there is no "true" class
                    # need to ignore this later
                    pass
                else:
                    # set the class probability
                    one_hot_label[batch_index, class_index, target_index] += (
                        1.0 - self.smoothing_value
                    )

        return one_hot_label

    def forward(
        self, predictions: torch.Tensor, labels: torch.Tensor, size_average=True
    ):
        """
        Calculate k-dimensional cross entropy loss
        :param predictions: shape: (batch, polarities e.g. 3, targets e.g. 5)
        :param labels: shape: (batch, targets)
        :param size_average:
        :return:
        prediction size earlier was batch polarities
        """
        batch_size, num_classes, num_targets = predictions.size()
        # calculate log of probability of each class (of each batch and target)
        pre_logsoftmax = self.logSoftmax(predictions)

        smooth_one_hot_for_true_targets = self._create_smooth_one_hot_for_true_targets(
            labels, batch_size, num_classes, num_targets
        ).to(self.device)

        # ignore those comparison where labels has a fillup value (=to be ignored)
        mask_nonfillup = labels != SentimentClasses.FILLUP_POLARITY_VALUE
        # shape is batch, targets
        # shape should be batch, classes, targets
        mask_nonfillup = mask_nonfillup.unsqueeze(1).repeat(1, num_classes, 1)
        # convert to 1 for True and 0 for False
        mask_nonfillup = mask_nonfillup.to(torch.int)
        # multiply so that scalars to be ignored are set to 0 (resulting in 0 loss for
        # those scalars, i.e., targets)
        # notes: categorical cross entropy loss does not directly punish on a low level
        # those predictions (scalars) that belong to incorrect classes (defined by true
        # or here, labels) but is only calculated by comparing the one true class (
        # defined by true, or here, labels) where it has a 1 (one hot encoded). since
        # the probability is 100% of all classes (also the output of the neural network)
        # the loss still punishes wrong predictions, i.e., if the class probability
        # should be 100% but is only 25% or 70%, the loss will be non-zero)
        # as a consequence, when there is no right class in "true", there cannot be a
        # loss. so, the multiplication below, which sets all fillup-targets to 0, has
        # the expected effect (no loss can result from fill up values, as all their
        # class probabilities are set to 0)
        smooth_one_hot_for_true_targets = (
            smooth_one_hot_for_true_targets * mask_nonfillup
        )

        # multiply
        loss = -smooth_one_hot_for_true_targets * pre_logsoftmax

        # aggregate loss to scalar over classes
        loss = torch.sum(loss, dim=1)
        # aggregate loss to scalar over targets
        loss = torch.sum(loss, dim=1)

        if size_average:
            return torch.mean(loss)
        else:
            return torch.sum(loss)



# adapted from absa-pytorch
from argparse import Namespace
from typing import Dict

import torch
import torch.nn as nn

from NewsSentiment.dataset import FXDataset
from NewsSentiment.layers.attention import Attention
from NewsSentiment.layers.point_wise_feed_forward import PositionwiseFeedForward
from NewsSentiment.layers.squeeze_embedding import SqueezeEmbedding

from NewsSentiment.consts import *
from NewsSentiment.fxlogger import get_logger
from NewsSentiment.models.FXBaseModel import FXBaseModel

logger = get_logger()


class AEN_Base(FXBaseModel):
    @staticmethod
    def get_language_models():
        return (BERT_BASE_UNCASED,)

    @staticmethod
    def get_input_field_ids():
        return [
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS),
            (BERT_BASE_UNCASED, FIELD_TARGET_IDS_WITH_SPECIAL_TOKENS),
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK),
        ]

    def __init__(self, transformer_models: Dict, opt: Namespace):
        super(AEN_Base, self).__init__()
        logger.info("creating AEN_Base")
        self.device = opt.device

        self.language_model = transformer_models[BERT_BASE_UNCASED]
        self.name = "aen_bert"
        self.lm_representation = "last"
        embed_dim = self.language_model.config.hidden_size

        self.squeeze_embedding = SqueezeEmbedding()
        self.dropout = nn.Dropout(opt.dropout)
        hidden_dim = embed_dim  # or should this be 300, as mentioned in the paper

        self.attn_k = Attention(
            embed_dim,
            out_dim=hidden_dim,
            n_head=8,
            score_function="mlp",
            dropout=opt.dropout,
        )
        self.attn_q = Attention(
            embed_dim,
            out_dim=hidden_dim,
            n_head=8,
            score_function="mlp",
            dropout=opt.dropout,
        )
        self.ffn_c = PositionwiseFeedForward(hidden_dim, dropout=opt.dropout)
        self.ffn_t = PositionwiseFeedForward(hidden_dim, dropout=opt.dropout)

        self.attn_s1 = Attention(
            hidden_dim, n_head=8, score_function="mlp", dropout=opt.dropout
        )

        self.dense = nn.Linear(hidden_dim * 3, opt.polarities_dim)

    def apply_lm(self, _input, _input_attention=None):
        if self.name in ["aen_bert", "aen_roberta"]:
            last_hidden, _, all_hidden = self.language_model(
                input_ids=_input, attention_mask=_input_attention
            )
        elif self.name == "aen_distilbert":
            last_hidden, all_hidden = self.language_model(
                input_ids=_input, attention_mask=_input_attention
            )
        else:
            raise Exception("unknown model name")

        if self.lm_representation == "last":
            return last_hidden
        elif self.lm_representation == "sum_last_four":
            last_four = all_hidden[-4:]  # list of four, each has shape: 16, 80, 768
            last_four_stacked = torch.stack(last_four)  # shape: 4, 16, 80, 768
            sum_last_four = torch.sum(last_four_stacked, dim=0)
            return sum_last_four
        elif self.lm_representation == "mean_last_four":
            last_four = all_hidden[-4:]  # list of four, each has shape: 16, 80, 768
            last_four_stacked = torch.stack(last_four)  # shape: 4, 16, 80, 768
            mean_last_four = torch.mean(last_four_stacked, dim=0)
            return mean_last_four
        elif self.lm_representation == "sum_last_two":
            last_two = all_hidden[-2:]
            last_two_stacked = torch.stack(last_two)
            sum_last_two = torch.sum(last_two_stacked, dim=0)
            return sum_last_two
        elif self.lm_representation == "mean_last_two":
            last_two = all_hidden[-2:]
            last_two_stacked = torch.stack(last_two)
            mean_last_two = torch.mean(last_two_stacked, dim=0)
            return mean_last_two
        elif self.lm_representation == "sum_all":
            all_stacked = torch.stack(all_hidden)
            sum_all = torch.sum(all_stacked, dim=0)
            return sum_all
        elif self.lm_representation == "mean_all":
            all_stacked = torch.stack(all_hidden)
            mean_all = torch.mean(all_stacked, dim=0)
            return mean_all

    def forward(self, inputs):
        context = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS
        )
        target = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TARGET_IDS_WITH_SPECIAL_TOKENS
        )
        context_len = torch.sum(context != 0, dim=-1)
        target_len = torch.sum(target != 0, dim=-1)

        context = self.squeeze_embedding(context, context_len)
        # context_attention = self.squeeze_embedding(context_attention, context_len)
        context = self.apply_lm(context)
        context = self.dropout(context)

        target = self.squeeze_embedding(target, target_len)
        # target_attention = self.squeeze_embedding(target_attention, target_len)
        target = self.apply_lm(target)
        target = self.dropout(target)

        hc, _ = self.attn_k(context, context)
        hc = self.ffn_c(hc)

        ht, _ = self.attn_q(context, target)
        ht = self.ffn_t(ht)

        s1, _ = self.attn_s1(hc, ht)

        context_len = torch.tensor(context_len, dtype=torch.float).to(self.device)
        target_len = torch.tensor(target_len, dtype=torch.float).to(self.device)

        hc_mean = torch.div(
            torch.sum(hc, dim=1), context_len.view(context_len.size(0), 1)
        )
        ht_mean = torch.div(
            torch.sum(ht, dim=1), target_len.view(target_len.size(0), 1)
        )
        s1_mean = torch.div(
            torch.sum(s1, dim=1), context_len.view(context_len.size(0), 1)
        )

        x = torch.cat((hc_mean, s1_mean, ht_mean), dim=-1)
        out = self.dense(x)

        return out



# -*- coding: utf-8 -*-
# file: attention.py
# author: songyouwei <youwei0314@gmail.com> for top two classes, huggingface for third (slightly
# adapted by FH)
# Copyright (C) 2020. All Rights Reserved.

import math
import torch
import torch.nn as nn
import torch.nn.functional as F


class Attention(nn.Module):
    def __init__(
        self,
        embed_dim,
        hidden_dim=None,
        out_dim=None,
        n_head=1,
        score_function="dot_product",
        dropout=0,
    ):
        """ Attention Mechanism
        :param embed_dim:
        :param hidden_dim:
        :param out_dim:
        :param n_head: num of head (Multi-Head Attention)
        :param score_function: scaled_dot_product / mlp (concat) / bi_linear (general dot)
        :return (?, q_len, out_dim,)
        """
        super(Attention, self).__init__()
        if hidden_dim is None:
            hidden_dim = embed_dim // n_head
        if out_dim is None:
            out_dim = embed_dim
        self.embed_dim = embed_dim
        self.hidden_dim = hidden_dim
        self.n_head = n_head
        self.score_function = score_function
        self.w_k = nn.Linear(embed_dim, n_head * hidden_dim)
        self.w_q = nn.Linear(embed_dim, n_head * hidden_dim)
        self.proj = nn.Linear(n_head * hidden_dim, out_dim)
        self.dropout = nn.Dropout(dropout)
        if score_function == "mlp":
            self.weight = nn.Parameter(torch.Tensor(hidden_dim * 2))
        elif self.score_function == "bi_linear":
            self.weight = nn.Parameter(torch.Tensor(hidden_dim, hidden_dim))
        else:  # dot_product / scaled_dot_product
            self.register_parameter("weight", None)
        self.reset_parameters()

    def reset_parameters(self):
        stdv = 1.0 / math.sqrt(self.hidden_dim)
        if self.weight is not None:
            self.weight.data.uniform_(-stdv, stdv)

    def forward(self, k, q):
        if len(q.shape) == 2:  # q_len missing
            q = torch.unsqueeze(q, dim=1)
        if len(k.shape) == 2:  # k_len missing
            k = torch.unsqueeze(k, dim=1)
        mb_size = k.shape[0]  # ?
        k_len = k.shape[1]
        q_len = q.shape[1]
        # k: (?, k_len, embed_dim,)
        # q: (?, q_len, embed_dim,)
        # kx: (n_head*?, k_len, hidden_dim)
        # qx: (n_head*?, q_len, hidden_dim)
        # score: (n_head*?, q_len, k_len,)
        # output: (?, q_len, out_dim,)
        kx = self.w_k(k).view(mb_size, k_len, self.n_head, self.hidden_dim)
        kx = kx.permute(2, 0, 1, 3).contiguous().view(-1, k_len, self.hidden_dim)
        qx = self.w_q(q).view(mb_size, q_len, self.n_head, self.hidden_dim)
        qx = qx.permute(2, 0, 1, 3).contiguous().view(-1, q_len, self.hidden_dim)
        if self.score_function == "dot_product":
            kt = kx.permute(0, 2, 1)
            score = torch.bmm(qx, kt)
        elif self.score_function == "scaled_dot_product":
            kt = kx.permute(0, 2, 1)
            qkt = torch.bmm(qx, kt)
            score = torch.div(qkt, math.sqrt(self.hidden_dim))
        elif self.score_function == "mlp":
            kxx = torch.unsqueeze(kx, dim=1).expand(-1, q_len, -1, -1)
            qxx = torch.unsqueeze(qx, dim=2).expand(-1, -1, k_len, -1)
            kq = torch.cat((kxx, qxx), dim=-1)  # (n_head*?, q_len, k_len, hidden_dim*2)
            score = torch.tanh(torch.matmul(kq, self.weight))
        elif self.score_function == "bi_linear":
            qw = torch.matmul(qx, self.weight)
            kt = kx.permute(0, 2, 1)
            score = torch.bmm(qw, kt)
        else:
            raise RuntimeError("invalid score_function")
        score = F.softmax(score, dim=-1)
        output = torch.bmm(score, kx)  # (n_head*?, q_len, hidden_dim)
        output = torch.cat(
            torch.split(output, mb_size, dim=0), dim=-1
        )  # (?, q_len, n_head*hidden_dim)
        output = self.proj(output)  # (?, q_len, out_dim)
        output = self.dropout(output)
        return output, score


class NoQueryAttention(Attention):
    """q is a parameter"""

    def __init__(
        self,
        embed_dim,
        hidden_dim=None,
        out_dim=None,
        n_head=1,
        score_function="dot_product",
        q_len=1,
        dropout=0,
    ):
        super(NoQueryAttention, self).__init__(
            embed_dim, hidden_dim, out_dim, n_head, score_function, dropout
        )
        self.q_len = q_len
        self.q = nn.Parameter(torch.Tensor(q_len, embed_dim))
        self.reset_q()

    def reset_q(self):
        stdv = 1.0 / math.sqrt(self.embed_dim)
        self.q.data.uniform_(-stdv, stdv)

    def forward(self, k, **kwargs):
        mb_size = k.shape[0]
        q = self.q.expand(mb_size, -1, -1)
        return super(NoQueryAttention, self).forward(k, q)


class FXBertSelfAttention(nn.Module):
    """
    Identical to the class BertSelfAttention by transformers by huggingface, but
    can be used without the need to have a config instance of class BertConfig but
    instead the required values can be passed on directly
    """
    def __init__(self, hidden_size, num_attention_heads, attention_probs_dropout_prob=0.1):
        super().__init__()
        if hidden_size % num_attention_heads != 0:
            raise ValueError(
                "The hidden size (%d) is not a multiple of the number of attention "
                "heads (%d)" % (hidden_size, num_attention_heads)
            )

        self.num_attention_heads = num_attention_heads
        self.attention_head_size = int(hidden_size / num_attention_heads)
        self.all_head_size = self.num_attention_heads * self.attention_head_size

        self.query = nn.Linear(hidden_size, self.all_head_size)
        self.key = nn.Linear(hidden_size, self.all_head_size)
        self.value = nn.Linear(hidden_size, self.all_head_size)

        self.dropout = nn.Dropout(attention_probs_dropout_prob)

    def transpose_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (self.num_attention_heads, self.attention_head_size)
        x = x.view(*new_x_shape)
        return x.permute(0, 2, 1, 3)

    def forward(
            self,
            hidden_states,
            attention_mask=None,
            head_mask=None,
            encoder_hidden_states=None,
            encoder_attention_mask=None,
            output_attentions=False,
    ):
        mixed_query_layer = self.query(hidden_states)

        # If this is instantiated as a cross-attention module, the keys
        # and values come from an encoder; the attention mask needs to be
        # such that the encoder's padding tokens are not attended to.
        if encoder_hidden_states is not None:
            mixed_key_layer = self.key(encoder_hidden_states)
            mixed_value_layer = self.value(encoder_hidden_states)
            attention_mask = encoder_attention_mask
        else:
            mixed_key_layer = self.key(hidden_states)
            mixed_value_layer = self.value(hidden_states)

        query_layer = self.transpose_for_scores(mixed_query_layer)
        key_layer = self.transpose_for_scores(mixed_key_layer)
        value_layer = self.transpose_for_scores(mixed_value_layer)

        # Take the dot product between "query" and "key" to get the raw attention scores.
        attention_scores = torch.matmul(query_layer, key_layer.transpose(-1, -2))
        attention_scores = attention_scores / math.sqrt(self.attention_head_size)
        if attention_mask is not None:
            # Apply the attention mask is (precomputed for all layers in BertModel forward() function)
            attention_scores = attention_scores + attention_mask

        # Normalize the attention scores to probabilities.
        attention_probs = nn.Softmax(dim=-1)(attention_scores)

        # This is actually dropping out entire tokens to attend to, which might
        # seem a bit unusual, but is taken from the original Transformer paper.
        attention_probs = self.dropout(attention_probs)

        # Mask heads if we want to
        if head_mask is not None:
            attention_probs = attention_probs * head_mask

        context_layer = torch.matmul(attention_probs, value_layer)

        context_layer = context_layer.permute(0, 2, 1, 3).contiguous()
        new_context_layer_shape = context_layer.size()[:-2] + (self.all_head_size,)
        context_layer = context_layer.view(*new_context_layer_shape)

        outputs = (context_layer, attention_probs) if output_attentions else (context_layer,)
        return outputs


import logging

logger = None


def get_logger():
    global logger

    if not logger:
        formatter = logging.Formatter(fmt='%(asctime)s %(levelname)s %(module)s(%(lineno)d):%(funcName)s|: %(message)s',
                                      datefmt='%H:%M:%S')

        handler = logging.StreamHandler()
        handler.setFormatter(formatter)

        logger = logging.getLogger("NewsSentiment")
        if logger.level == logging.NOTSET:
            logger.setLevel(logging.ERROR)

        if len(logger.handlers) == 0:
            logger.addHandler(handler)

    return logger



from argparse import Namespace
from typing import Dict

import torch
import torch.nn as nn

from NewsSentiment.consts import *
from NewsSentiment.dataset import FXDataset, FXEasyTokenizer
from NewsSentiment.models.FXBaseModel import FXBaseModel, provide_pretrained, default_pretrained


@default_pretrained("v1.0.0")
@provide_pretrained("v1.0.0", "https://github.com/fhamborg/NewsMTSC/releases/download/v1.0.0/grutsc")
class GRUTSCSingle(FXBaseModel):
    """
    Inspired from https://arxiv.org/pdf/2006.00052.pdf
    Differences:
    - instead of question ("Is the ACLU good for USA?") then text (1 or more sentences),
      we use text then target (and no question, similar to BERT-SPC)
    - no vader
    - additionally we can flexibly use any knowledge source as well as multiple
    - we have one large matrix for all concatenated knowledge source embeddings, whereas
      in the original paper they use individual, smaller matrices for each knowledge
      source embedding
    - target mask (mostly useful for BERT)
    - fine-tuning LM enabled
    """

    @staticmethod
    def get_language_models():
        return (get_default_lm(),)

    @staticmethod
    def get_input_field_ids():
        return [
            (get_default_lm(), FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS),
            (
                get_default_lm(),
                FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS_SEGMENT_IDS,
            ),
            # while we use text-then-target as bert input, we can use text targetmask
            # and text knowledge source mask because it is identical to a hypothetical
            # text-then-target target mask or text-then-target knowledge source mask
            # (we would not highlight the target in the 2nd component in the
            # corresponding mask)
            (get_default_lm(), FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK),
            (
                get_default_lm(),
                FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_SELECTED_KNOWLEDGE_SOURCES,
            ),
        ]

    def __init__(self, transformer_models: Dict, opt: Namespace):
        super().__init__()
        self.language_model = transformer_models[get_default_lm()]
        self.ks_embeddings_dense = nn.Linear(
            FXEasyTokenizer.NUM_CATEGORIES_OF_SELECTED_KNOWLEDGE_SOURCES,
            self.language_model.config.hidden_size,
        )
        # if get_default_lm() == BERT_BASE_UNCASED:
        #     self.is_use_targetmask = True
        # else:
        #     self.is_use_targetmask = False
        self.is_use_targetmask = False
        num_input_embeddings = 2
        if self.is_use_targetmask:
            num_input_embeddings = 3

        self.gru = nn.GRU(
            self.language_model.config.hidden_size * num_input_embeddings,
            self.language_model.config.hidden_size * num_input_embeddings,
            bidirectional=True,
            batch_first=True,
        )
        self.dropout = nn.Dropout(opt.dropout)
        self.dense = nn.Linear(
            # 3 inputs (original last gru out, mean, max), 2 inputs to gru (bert and
            # knowledge embedding), 2 (because bidirectional gru)
            self.language_model.config.hidden_size * 3 * num_input_embeddings * 2,
            opt.polarities_dim,
        )
        self.device = opt.device

    def forward(self, inputs, is_return_ensemble_values: bool = False):
        # get inputs
        text_target_bert_indices = FXDataset.get_input_by_params(
            inputs, get_default_lm(), FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS
        )
        text_target_bert_segments_ids = FXDataset.get_input_by_params(
            inputs,
            get_default_lm(),
            FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS_SEGMENT_IDS,
        )
        text_bert_indices_target_mask = FXDataset.get_input_by_params(
            inputs, get_default_lm(), FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK
        )
        text_bert_indices_selected_knowledge_sources = FXDataset.get_input_by_params(
            inputs,
            get_default_lm(),
            FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_SELECTED_KNOWLEDGE_SOURCES,
        )

        # apply bert
        last_hidden_states = self.invoke_language_model(
            lm=self.language_model,
            input_ids=text_target_bert_indices,
            token_type_ids=text_target_bert_segments_ids,
        )
        # shape: batch, seqlen, hiddendim

        # apply knowledge embedding
        knowledge_embedded = self.ks_embeddings_dense(
            text_bert_indices_selected_knowledge_sources.float()
        )
        # shape: batch, seqlen, hiddendim

        if self.is_use_targetmask:
            # repeat
            target_mask = text_bert_indices_target_mask.unsqueeze(dim=2).repeat(
                1, 1, knowledge_embedded.shape[2]
            )
            # shape: batch, seqlen, hiddendim

            # concat (called x_t in paper)
            bert_and_knowledge = torch.cat(
                (last_hidden_states, knowledge_embedded, target_mask), dim=2
            )
            # batch x seq x bert+knowledge+targetmask
        else:
            # concat (called x_t in paper)
            bert_and_knowledge = torch.cat(
                (last_hidden_states, knowledge_embedded), dim=2
            )
            # batch x seq x bert+knowledge

        # apply gru (result called z_t in paper)
        gru_all_hidden, gru_last_hidden = self.gru(
            bert_and_knowledge,
            torch.zeros(
                2,
                bert_and_knowledge.shape[0],
                self.language_model.config.hidden_size * 2,
            ).to(self.device),
        )
        # all hidden shape: batch x seq x 4*hidden (contains hidden states for each
        # part of the input seq)
        # last hidden shap: numdir x batch x 2*hidden (contains hidden states for last
        # part of input seq)

        # gru_last_hidden_own = gru_all_hidden[:,-1:,]
        # get both directions
        gru_last_hidden_dir0 = gru_last_hidden[0, :, :]
        gru_last_hidden_dir1 = gru_last_hidden[1, :, :]
        # shape each: batch x 2*hidden
        gru_last_hidden_stacked = torch.cat(
            (gru_last_hidden_dir0, gru_last_hidden_dir1), dim=1
        )
        # batch x 4*hidden

        # pooling
        # according to original paper: "max-pooling returns a vector with maximum
        # weights across all hidden states of input tokens for each dimension. in this
        # way, the input tokens with higher weights will be engaged for stance
        # prediction."
        gru_avg = torch.mean(gru_all_hidden, dim=1)
        gru_max, _ = torch.max(gru_all_hidden, dim=1)

        # concat (called "u" in original paper)
        gru_complete_concatted = torch.cat(
            (gru_last_hidden_stacked, gru_avg, gru_max), dim=1
        )

        if is_return_ensemble_values:
            return gru_complete_concatted
        else:
            # dense
            logits = self.dense(gru_complete_concatted)

            return logits



import torch.nn as nn
import torch


class AggregatorForBert(nn.Module):
    """
    According to https://huggingface.co/transformers/model_doc/bert.html#bertmodel we
    should not (as in the original SPC version) the pooler_output but get the
    last_hidden_state and "averaging or pooling the sequence of hidden-states for the
    whole input sequence"
    """
    def __init__(self, spc_lm_representation: str):
        super(AggregatorForBert, self).__init__()
        self.spc_lm_representation = spc_lm_representation

    def forward(self, last_hidden_state, pooler_output, all_hidden_states):
        if self.spc_lm_representation == "pooler_output":
            prepared_output = pooler_output
        elif self.spc_lm_representation == "mean_last":
            prepared_output = last_hidden_state.mean(dim=1)
        elif self.spc_lm_representation == "mean_last_four":
            prepared_output = (
                torch.stack(all_hidden_states[-4:]).mean(dim=0).mean(dim=1)
            )
        elif self.spc_lm_representation == "mean_last_two":
            prepared_output = (
                torch.stack(all_hidden_states[-2:]).mean(dim=0).mean(dim=1)
            )
        elif self.spc_lm_representation == "mean_all":
            prepared_output = torch.stack(all_hidden_states).mean(dim=0).mean(dim=1)
        elif self.spc_lm_representation == "sum_last":
            prepared_output = last_hidden_state.sum(dim=1)
        elif self.spc_lm_representation == "sum_last_four":
            prepared_output = torch.stack(all_hidden_states[-4:]).sum(dim=0).sum(dim=1)
        elif self.spc_lm_representation == "sum_last_two":
            prepared_output = torch.stack(all_hidden_states[-2:]).sum(dim=0).sum(dim=1)
        elif self.spc_lm_representation == "sum_all":
            prepared_output = torch.stack(all_hidden_states).sum(dim=0).sum(dim=1)
        return prepared_output



from argparse import Namespace
from typing import Dict

import torch.nn as nn

from NewsSentiment.consts import *
from NewsSentiment.dataset import FXDataset
from NewsSentiment.models.FXBaseModel import FXBaseModel


class TDBertLikeMultiTarget(FXBaseModel):
    """
    This model returns a sequence that only contains the hidden states of those output
    nodes that represent a word piece of the target phrase. All other hidden states are
    set to 0. The length of the output sequence is opt.max_seq_len

    From a conceptual perspective, TD-BERT is in some aspects similar (e.g., only the
    output of target-phrase-related nodes is used), but in other aspects not similar or
    it is not quite clear how the authors of TD-BERT implemented them. An email I sent
    to them was not answered yet.
    """

    @staticmethod
    def get_language_models():
        return (BERT_BASE_UNCASED,)

    @staticmethod
    def get_input_field_ids():
        return [
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS),
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK),
            # (
            #     BERT_BASE_UNCASED,
            #     FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_SELECTED_KNOWLEDGE_SOURCES,
            # ),
        ]

    def __init__(self, transformer_models: Dict, opt: Namespace):
        super(TDBertLikeMultiTarget, self).__init__()
        self.language_model = transformer_models[BERT_BASE_UNCASED]
        self.dropout = nn.Dropout(opt.dropout)
        self.dense = nn.Linear(
            self.language_model.config.hidden_size, opt.polarities_dim
        )

    def forward(self, inputs):
        # get inputs
        text_bert_indices = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS
        )
        text_bert_indices_targets_mask = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK
        )
        # text_bert_indices_nrc_emolex = FXDataset.get_input_by_params(
        #     inputs,
        #     BERT_BASE_UNCASED,
        #     FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_SELECTED_KNOWLEDGE_SOURCES,
        # )
        # prepare inputs
        # for text only, we do not need target specific information, i.e., all text
        # vectors are identical. also, bert, can
        # only process one sequence of size max_seq_len (more specifically, a tensor
        # of size batch_size x max_seq_ken). thus, we select only the first element from
        # the second dimension (the dimensions are: batch, targets, hidden_states)
        text_bert_indices = text_bert_indices[:, 0, :]
        # apply bert
        last_hidden_states, pooler_output, all_hidden_states = self.language_model(
            input_ids=text_bert_indices
        )
        # element-wise multiplication with target mask
        # align the dimensions of the tensors
        # last_hidden_states.shape = 4,150,768; should be 4,5,150,768
        # insert a new singleton dimension after the first dimension
        # new shape: 4,1,150,768
        last_hidden_states = last_hidden_states.unsqueeze(1)
        # repeat k times along the new 2nd dimension, where k is the target size
        last_hidden_states = last_hidden_states.repeat(
            1, text_bert_indices_targets_mask.shape[1], 1, 1
        )
        # text_bert_indices_targets_mask.shape = 4,5,150; should be 4,5,150,768
        # insert singleton simension after the three already existing dimensions
        text_bert_indices_targets_mask = text_bert_indices_targets_mask.unsqueeze(3)
        text_bert_indices_targets_mask = text_bert_indices_targets_mask.repeat(
            1, 1, 1, last_hidden_states.shape[3]
        )

        last_hidden_states_only_targets = (
            last_hidden_states * text_bert_indices_targets_mask
        )

        # similar to TD-BERT, perform max pooling TODO not implemented yet, instead:
        # for now, retrieve only the values of the target's output tokens and then
        # calculate the mean:
        # (batchsize, targetsize, 150, 768) -> (batchsize, targetsize, 768)
        # get the positions of target nodes. note that we cannot simply take the mean
        # as it would divide by the number of the sequence length, whereas the effective
        # length is only of size k, where k is the number of non-zero scalars in the
        # input mask (since we are only interested in those values)
        last_hidden_states_aggregated_per_target = last_hidden_states_only_targets.sum(
            dim=2
        )
        # get the sum for each (batch, hidden states) (sum over the sequence length
        # dim)
        denominator_for_mean = text_bert_indices_targets_mask.sum(dim=2)
        # divide for each (batch, hidden states) by the denominator to get the mean
        last_hidden_states_aggregated_per_target = (
            last_hidden_states_aggregated_per_target / denominator_for_mean
        )
        # dropout before dense layer, as in most other tsc models
        last_hidden_states_aggregated_per_target = self.dropout(
            last_hidden_states_aggregated_per_target
        )
        # dense layer
        logits = self.dense(last_hidden_states_aggregated_per_target)

        return logits



# this file re-implements TD-BERT-QA by Gao Zhengjie et al.
from argparse import Namespace
from typing import Dict

import torch
import torch.nn as nn
import torch.nn.functional as F

from NewsSentiment.consts import *
from NewsSentiment.dataset import FXDataset
from NewsSentiment.models.FXBaseModel import FXBaseModel


class TD_BERT_QA_MUL(FXBaseModel):
    @staticmethod
    def get_language_models():
        return (BERT_BASE_UNCASED,)

    @staticmethod
    def get_input_field_ids():
        return [
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS),
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK),
        ]

    def __init__(self, transformer_models: Dict, opt: Namespace):
        super(TD_BERT_QA_MUL, self).__init__()
        self.opt = opt
        self.language_model = transformer_models[BERT_BASE_UNCASED]
        self.dropout = nn.Dropout(opt.dropout)
        self.fc = nn.Linear(self.language_model.config.hidden_size, opt.polarities_dim)  # 全连接层 bbfc
        self.bn = nn.BatchNorm1d(self.language_model.config.hidden_size)

    def forward(self, inputs):
        # get inputs
        text_bert_indices = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS
        )
        text_bert_indices_target_mask = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK
        )

        # apply bert
        last_hidden_states, pooler_output, all_hidden_states = self.language_model(
            input_ids=text_bert_indices
        )

        # element-wise multiplication with target mask
        # unsqueeze, cf. https://stackoverflow.com/q/62559382
        text_bert_indices_target_mask_unsqueezed = text_bert_indices_target_mask.unsqueeze(
            -1
        )
        last_hidden_states_only_target = (
                last_hidden_states * text_bert_indices_target_mask_unsqueezed
        )

        # as in TD-BERT, perform max pooling
        last_hidden_states_only_target_aggregated, _ = last_hidden_states_only_target.max(
            dim=1
        )

        target_in_sent_embed = self.bn(last_hidden_states_only_target_aggregated)
        target_in_sent_embed = target_in_sent_embed.mul(pooler_output)
        cat = self.dropout(target_in_sent_embed)

        logits = self.fc(cat)
        # removed tanh, which was invoked in original tdbert. for training, we dont
        # need it to properly compute the loss. we would, however, need softmax during
        # inferring to have the probabilities of all mutually exclusive classes
        # to sum up to 1

        return logits


class TD_BERT_QA_CON(FXBaseModel):
    @staticmethod
    def get_language_models():
        return (BERT_BASE_UNCASED,)

    @staticmethod
    def get_input_field_ids():
        return [
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS),
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK),
        ]

    def __init__(self, transformer_models: Dict, opt: Namespace):
        super(TD_BERT_QA_CON, self).__init__()
        self.opt = opt
        self.language_model = transformer_models[BERT_BASE_UNCASED]
        self.dropout = nn.Dropout(opt.dropout)
        self.fc = nn.Linear(self.language_model.config.hidden_size*2, opt.polarities_dim)  # 全连接层 bbfc
        self.bn = nn.BatchNorm1d(self.language_model.config.hidden_size)

    def forward(self, inputs):
        # get inputs
        text_bert_indices = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS
        )
        text_bert_indices_target_mask = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK
        )

        # apply bert
        last_hidden_states, pooler_output, all_hidden_states = self.language_model(
            input_ids=text_bert_indices
        )

        # element-wise multiplication with target mask
        # unsqueeze, cf. https://stackoverflow.com/q/62559382
        text_bert_indices_target_mask_unsqueezed = text_bert_indices_target_mask.unsqueeze(
            -1
        )
        last_hidden_states_only_target = (
                last_hidden_states * text_bert_indices_target_mask_unsqueezed
        )

        # as in TD-BERT, perform max pooling
        last_hidden_states_only_target_aggregated, _ = last_hidden_states_only_target.max(
            dim=1
        )

        # not entirely sure whether this is as in original tdbertqa-con, because the code does not exist in the repo
        # (seems to be part of the commented lines in there)
        pooler_output = self.bn(pooler_output)

        cat = torch.cat([pooler_output, last_hidden_states_only_target_aggregated], dim=1)
        cat = self.dropout(cat)
        logits = self.fc(cat)
        # removed tanh, which was invoked in original tdbert. for training, we dont
        # need it to properly compute the loss. we would, however, need softmax during
        # inferring to have the probabilities of all mutually exclusive classes
        # to sum up to 1

        return logits



import argparse

import torch
import torch.nn.functional as F
from jsonlines import jsonlines
from tqdm import tqdm

from NewsSentiment.SentimentClasses import SentimentClasses
from NewsSentiment.dataset import FXEasyTokenizer
from NewsSentiment.download import Download
from NewsSentiment.fxlogger import get_logger
from NewsSentiment.models.singletarget.grutscsingle import GRUTSCSingle
from NewsSentiment.train import (
    parse_arguments as parse_arguments_from_train,
    prepare_and_start_instructor,
)


class TargetSentimentClassifier:
    def __init__(
        self,
        opts_from_infer=None,
        single_targets=True,
        logging_level="ERROR",
        state_dict=None,
    ):
        self.logger = get_logger()

        # get the default parameters from train.py
        final_opts = parse_arguments_from_train(
            override_args=False, overwrite_logging_level=logging_level
        )
        # if no arguments have been passed, get the defaults of infer.py (and since
        # override_args=True, not the arguments passed by the console (in args) will be
        # used but just the defaults defined in parse_arguments
        if not opts_from_infer:
            opts_from_infer = parse_arguments(override_args=True)

        for key, val in vars(opts_from_infer).items():
            if val is not None:
                previous_val = getattr(final_opts, key)
                self.logger.info("overwriting: %s=%s to %s", key, previous_val, val)
                setattr(final_opts, key, val)

        # set custom variables
        final_opts.single_targets = single_targets
        final_opts.multi_targets = not single_targets
        if state_dict:
            final_opts.state_dict = state_dict

        # set training_mode to False so that we get the Instructor object
        final_opts.training_mode = False

        # prepare and initialize instructor
        instructor = prepare_and_start_instructor(final_opts)

        # get stuff that we need from the instructor
        self.model = instructor.own_model
        self.tokenizer = instructor.own_tokenizer
        self.opt = instructor.opt
        self.instructor = instructor
        self.polarities_inverse = SentimentClasses.get_polarity_associations_inverse()

        # set model to evaluation mode (disable gradient / learning)
        self.model.eval()
        torch.autograd.set_grad_enabled(False)

    def infer_from_text(self, left, target, right):
        """
        Calculates the targeted mention of the sentence defined by left + target + right.
        :param left: The text from the beginning of the sentence to the mention of the
        target (will be empty if the sentence starts with the mention of the target).
        Make sure to include a space if there is one between the left phrase and the
        target mention, e.g., "I like Peter." -> left="I like ", target="Peter", right=".".
        :param target: The mention of the target.
        :param right: The text from the end of the target to the end of the sentence.
        Make sure to include a space if there is one between the target and the next word
        of the sentence.
        :return:
        """
        return self.infer(text_left=left, target_mention=target, text_right=right)

    def infer(
        self,
        text_left: str = None,
        target_mention: str = None,
        text_right: str = None,
        text: str = None,
        target_mention_from: int = None,
        target_mention_to: int = None,
    ):
        """
        Calculates sentiment as to target_mention in a text that is a concatenation of
        text_left,
        target_mention, and text_right. Note that text_left and text_right should end
        with a space (or comma, etc.)),
        or end with a space, respectively. Alternatively, the target can be selected via
        target_mention_from and target_mention_to in text.
        """
        assert not text_left and text or text_left and not text

        if text:
            text_left = text[:target_mention_from]
            target_mention = text[target_mention_from:target_mention_to]
            text_right = text[target_mention_from:]

        # assert text_left.endswith(' ') # we cannot handle commas, if we have this
        # check
        assert not target_mention.startswith(" ") and not target_mention.endswith(
            " "
        ), f"target_mention={target_mention}; text={text}"
        # assert text_right.startswith(' ')

        text_left = FXEasyTokenizer.prepare_left_segment(text_left)
        target_mention = FXEasyTokenizer.prepare_target_mention(target_mention)
        text_right = FXEasyTokenizer.prepare_right_segment(text_right)

        indexed_example = self.tokenizer.create_model_input_seqs(
            text_left, target_mention, text_right, []
        )
        inputs = self.instructor.select_inputs(indexed_example, is_single_item=True)

        # invoke model
        outputs = self.model(inputs)
        class_probabilites = F.softmax(outputs, dim=-1).reshape((3,)).cpu().tolist()

        classification_result = []
        for class_id, class_prob in enumerate(class_probabilites):
            classification_result.append(
                {
                    "class_id": class_id,
                    "class_label": self.polarities_inverse[class_id],
                    "class_prob": class_prob,
                }
            )

        classification_result = sorted(
            classification_result, key=lambda x: x["class_prob"], reverse=True
        )

        return classification_result

    def get_info_for_label(self, classification_result, label):
        for r in classification_result:
            if r["class_label"] == label:
                return r
        raise ValueError(label)


def parse_arguments(override_args=False):
    parser = argparse.ArgumentParser()
    parser.add_argument("--own_model_name", default="grutsc", type=str)
    parser.add_argument(
        "--pretrained_model_name",
        type=str,
        help="has to be placed in folder pretrained_models",
    )
    parser.add_argument("--default_lm", default="roberta-base", type=str)
    parser.add_argument(
        "--state_dict", type=str, default=Download.model_path(GRUTSCSingle)
    )
    parser.add_argument(
        "--knowledgesources", default="nrc_emotions mpqa_subjectivity bingliu_opinion"
    )
    parser.add_argument(
        "--device",
        default=None,
        type=str,
        help="e.g., cuda:0; if None: any CUDA device will be used if available, else "
        "CPU",
    )
    parser.add_argument(
        "--logging", type=str, default="ERROR",
    )

    # if own_args == None -> parse_args will use sys.argv
    # if own_args == [] -> parse_args will use this empty list instead
    own_args = None
    if override_args:
        own_args = []

    # create arguments
    opt = parser.parse_args(args=own_args)

    return opt



"""
Download a specific version of a finetuned model and place it in pretrained_models.
"""
import os
import string

import torch

from NewsSentiment.fxlogger import get_logger


class Download:

    def __init__(
        self, own_model_name, version="default", force=False, list_versions=False
    ):
        from NewsSentiment.models.FXBaseModel import model_includes_pretrained
        from NewsSentiment.train import OWN_MODELNAME2CLASS

        logger = get_logger()
        own_model_name = own_model_name
        if version == "default":
            version = None
        try:
            model_cls = OWN_MODELNAME2CLASS[own_model_name]
        except KeyError:
            logger.error(f'Model "{own_model_name}" is unknown.')
            exit(2)
        if not model_includes_pretrained(model_cls):
            logger.error(
                f'Model "{own_model_name}" does not ship any pretrained models for download.'
            )
            exit(2)
        if list_versions:
            self.list_versions(model_cls, own_model_name)
        else:
            self.download(model_cls, version, force)

    @staticmethod
    def list_versions(model_cls, own_model_name=""):
        default = model_cls.get_pretrained_default_version()
        versions = model_cls.get_pretrained_versions()
        if own_model_name:
            own_model_name = f' "{own_model_name}"'
        print(f"Model{own_model_name} provides following pretrained versions:")
        for version, source in versions.items():
            default_str = ""
            if version == default:
                default_str = " (default)"
            print(f'"{version}"{default_str}: {source}')

    @classmethod
    def download(cls, model_cls, version=None, force=False, skip_if_exists=True):
        source = model_cls.get_pretrained_source(version)
        path = cls.model_path(model_cls, version)
        if not force and os.path.isfile(path):
            if skip_if_exists:
                return
            print("Model file already exists. Use --force to overwrite.")
            exit(2)
        print(f"Downloading to {path}:")
        os.makedirs(os.path.dirname(path), exist_ok=True)
        torch.hub.download_url_to_file(source, path)
        return path

    @staticmethod
    def model_filename(model_cls, version=None):
        version = version or model_cls.get_pretrained_default_version()
        source_filename = os.path.basename(model_cls.get_pretrained_source(version))
        name = f"{source_filename}_{version}"
        allowed = set(f".-_ {string.ascii_letters}{string.digits}")
        filename = "".join(char for char in name if char in allowed)
        return filename.replace(".", "-").replace(" ", "_")

    @classmethod
    def model_path(cls, model_cls, version=None):
        return os.path.join(torch.hub.get_dir(),
                            "pretrained_models",
                            "state_dicts",
                            cls.model_filename(model_cls, version))

    @staticmethod
    def add_subparser(subparser):
        """Download models for NewsSentiment"""
        subparser.add_argument("--own_model_name", default="grutsc", type=str)
        subparser.add_argument(
            "--version",
            default=None,
            type=str,
            help="version of the model to download, use --force to overwrite a version which was already downloaded",
        )
        subparser.add_argument(
            "--force",
            action="store_true",
            help="force the download of a model and overwrite potential previous versions",
        )
        subparser.add_argument(
            "--list_versions",
            action="store_true",
            help="List all pretrained model versions which a model provides",
        )
        return subparser

    @classmethod
    def run_from_parser(cls, args_namespace):
        args = vars(args_namespace)
        return cls(**args)




import os
from collections import defaultdict, Counter
from pathlib import Path
import csv

from NewsSentiment.fxlogger import get_logger
from NewsSentiment.diskdict import DiskDict

EMOTION2INDEX = {
    "anger": 0,
    "anticipation": 1,
    "disgust": 2,
    "fear": 3,
    "joy": 4,
    "negative": 5,
    "positive": 6,
    "sadness": 7,
    "surprise": 8,
    "trust": 9,
}
THIS_DIR = Path(os.path.dirname(os.path.realpath(__file__)))
PATH_DICT_NRC_EMOLEX = THIS_DIR / "NRC-Emotion-Lexicon-Wordlevel-v0.92.txt.ddict"

logger = get_logger()


def convert_txt_to_dict():
    path_dict = THIS_DIR / "NRC-Emotion-Lexicon-Wordlevel-v0.92.txt"

    term2emotions = defaultdict(set)
    emotion_counter = Counter()
    with open(path_dict, "r") as csv_file:
        csv_reader = csv.reader(csv_file, delimiter="\t")
        for term, emotion, is_present in csv_reader:
            if is_present == "1":
                emotion_index = EMOTION2INDEX[emotion]
                term2emotions[term].add(emotion_index)
                emotion_counter[emotion] += 1
            elif is_present == "0":
                pass
            else:
                raise ValueError
    logger.info("read %s terms", len(term2emotions))
    logger.info("emotion count:\n%s", emotion_counter.most_common())

    logger.info("saving to %s...", None)
    ddict_emolex = DiskDict(PATH_DICT_NRC_EMOLEX)
    ddict_emolex.update(term2emotions)
    ddict_emolex.sync_to_disk()
    logger.info("done")


if __name__ == "__main__":
    convert_txt_to_dict()



def find_key_original_or_lc(data: dict, data_keys_lower: dict, key: str):
    # try to get the key as it is from the dict
    if key in data.keys():
        return data[key]
    # if not contained, try whether if using case insensitivity we find an entry
    if key.lower() in data_keys_lower.keys():
        return data_keys_lower[key.lower()]
    # if not, return None
    return set()



# adapted from https://github.com/yangheng95/LCF-ABSA and
# https://github.com/StevePhan101/LCFS-BERT/
from argparse import Namespace
from typing import Dict

import numpy as np
import torch
import torch.nn as nn
from transformers.modeling_bert import BertPooler, BertSelfAttention

from NewsSentiment.consts import *
from NewsSentiment.dataset import FXDataset
from NewsSentiment.models.FXBaseModel import FXBaseModel


class PointwiseFeedForward(nn.Module):
    """
    A two-feed-forward-layer module
    """

    def __init__(self, d_hid, d_inner_hid=None, d_out=None, dropout=0):
        super(PointwiseFeedForward, self).__init__()
        if d_inner_hid is None:
            d_inner_hid = d_hid
        if d_out is None:
            d_out = d_inner_hid
        self.w_1 = nn.Conv1d(d_hid, d_inner_hid, 1)  # position-wise
        self.w_2 = nn.Conv1d(d_inner_hid, d_out, 1)  # position-wise
        self.dropout = nn.Dropout(dropout)
        self.relu = nn.ReLU()

    def forward(self, x):
        output = self.relu(self.w_1(x.transpose(1, 2)))
        output = self.w_2(output).transpose(2, 1)
        output = self.dropout(output)
        return output


class SelfAttention(nn.Module):
    def __init__(self, config, opt):
        super(SelfAttention, self).__init__()
        self.opt = opt
        self.config = config
        self.SA = BertSelfAttention(config)
        self.tanh = torch.nn.Tanh()

    def forward(self, inputs):
        zero_tensor = torch.tensor(
            np.zeros((inputs.size(0), 1, 1, self.opt.max_seq_len), dtype=np.float32),
            dtype=torch.float32,
        ).to(self.opt.device)
        SA_out = self.SA(inputs, zero_tensor)
        return self.tanh(SA_out[0])


class LCFS_BERT(FXBaseModel):
    @staticmethod
    def get_language_models():
        return (BERT_BASE_UNCASED,)

    @staticmethod
    def get_input_field_ids():
        return [
            (BERT_BASE_UNCASED, FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS),
            (
                BERT_BASE_UNCASED,
                FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS_SEGMENT_IDS,
            ),
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS),
            (BERT_BASE_UNCASED, FIELD_TARGET_IDS_WITH_SPECIAL_TOKENS),
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK),
            (BERT_BASE_UNCASED, FIELD_SYNTAX_HOP_DISTANCE_TO_TARGET),
        ]

    def __init__(self, transformer_models: Dict, opt: Namespace):
        super(LCFS_BERT, self).__init__()

        bert = transformer_models[BERT_BASE_UNCASED]
        self.bert_spc = bert
        self.opt = opt

        # self.bert_local = copy.deepcopy(bert)  # Uncomment the line to use dual Bert
        self.bert_local = (
            bert  # Default to use single Bert and reduce memory requirements
        )
        self.dropout = nn.Dropout(self.opt.dropout)
        self.bert_SA = SelfAttention(bert.config, self.opt)

        # self.linear_double = nn.Linear(
        #    bert.config.hidden_size * 2, bert.config.hidden_size
        # )
        self.mean_pooling_double = PointwiseFeedForward(
            bert.config.hidden_size * 2,
            bert.config.hidden_size,
            bert.config.hidden_size,
        )
        self.linear_single = nn.Linear(bert.config.hidden_size, bert.config.hidden_size)
        self.bert_pooler = BertPooler(bert.config)

        self.dense = nn.Linear(bert.config.hidden_size, self.opt.polarities_dim)

    def feature_dynamic_mask(self, text_local_indices, aspect_indices, distances_input):
        texts = text_local_indices.cpu().numpy()
        asps = aspect_indices.cpu().numpy()
        if distances_input is not None:
            distances_input = distances_input.cpu().numpy()
        mask_len = self.opt.SRD
        masked_text_raw_indices = np.ones(
            (
                text_local_indices.size(0),
                self.opt.max_seq_len,
                self.bert_local.config.hidden_size,
            ),
            dtype=np.float32,
        )
        for text_i, asp_i in zip(range(len(texts)), range(len(asps))):
            if distances_input is None:
                # this should never be reached
                asp_len = np.count_nonzero(asps[asp_i]) - 2
                try:
                    asp_begin = np.argwhere(texts[text_i] == asps[asp_i][1])[0][0]
                except:
                    continue
                if asp_begin >= mask_len:
                    mask_begin = asp_begin - mask_len
                else:
                    mask_begin = 0
                for i in range(mask_begin):
                    masked_text_raw_indices[text_i][i] = np.zeros(
                        (self.bert_local.config.hidden_size), dtype=np.float
                    )
                for j in range(asp_begin + asp_len + mask_len, self.opt.max_seq_len):
                    masked_text_raw_indices[text_i][j] = np.zeros(
                        (self.bert_local.config.hidden_size), dtype=np.float
                    )
                else:
                    distances_i = distances_input[text_i]
                for i, dist in enumerate(distances_i):
                    # iterate the distances and set the mask to 0 for those that have a
                    # too large distance
                    if dist > mask_len:
                        masked_text_raw_indices[text_i][i] = np.zeros(
                            (self.bert_local.config.hidden_size), dtype=np.float
                        )
        masked_text_raw_indices = torch.from_numpy(masked_text_raw_indices)
        return masked_text_raw_indices.to(self.opt.device)

    def feature_dynamic_weighted(
        self, text_local_indices, aspect_indices, distances_input
    ):
        texts = text_local_indices.cpu().numpy()
        asps = aspect_indices.cpu().numpy()
        if distances_input is not None:
            distances_input = distances_input.cpu().numpy()
        masked_text_raw_indices = np.ones(
            (
                text_local_indices.size(0),
                self.opt.max_seq_len,
                self.bert_local.config.hidden_size,
            ),
            dtype=np.float32,
        )
        mask_len = self.opt.SRD
        for text_i, asp_i in zip(range(len(texts)), range(len(asps))):
            if distances_input is None:
                asp_len = np.count_nonzero(asps[asp_i]) - 2
                try:
                    asp_begin = np.argwhere(texts[text_i] == asps[asp_i][1])[0][0]
                    asp_avg_index = (asp_begin * 2 + asp_len) / 2
                except:
                    continue
                distances = np.zeros(np.count_nonzero(texts[text_i]), dtype=np.float32)
                for i in range(1, np.count_nonzero(texts[text_i]) - 1):
                    if abs(i - asp_avg_index) + asp_len / 2 > self.opt.SRD:
                        distances[i] = 1 - (
                            abs(i - asp_avg_index) + asp_len / 2 - self.opt.SRD
                        ) / np.count_nonzero(texts[text_i])
                    else:
                        distances[i] = 1
                for i in range(len(distances)):
                    masked_text_raw_indices[text_i][i] = (
                        masked_text_raw_indices[text_i][i] * distances[i]
                    )
            else:
                distances_i = distances_input[text_i]  # distances of batch i-th
                for i, dist in enumerate(distances_i):
                    if dist > mask_len:
                        distances_i[i] = 1 - (dist - mask_len) / np.count_nonzero(
                            texts[text_i]
                        )
                    else:
                        distances_i[i] = 1

                for i in range(len(distances_i)):
                    masked_text_raw_indices[text_i][i] = (
                        masked_text_raw_indices[text_i][i] * distances_i[i]
                    )
        masked_text_raw_indices = torch.from_numpy(masked_text_raw_indices)
        return masked_text_raw_indices.to(self.opt.device)

    def forward(self, inputs):
        text_target_bert_indices = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS,
        )

        text_target_bert_segments_ids = FXDataset.get_input_by_params(
            inputs,
            BERT_BASE_UNCASED,
            FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS_SEGMENT_IDS,
        )
        text_local_indices = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS
        )
        aspect_indices = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TARGET_IDS_WITH_SPECIAL_TOKENS
        )
        syntax_hop_distance_to_target = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_SYNTAX_HOP_DISTANCE_TO_TARGET
        )

        bert_spc_out, _, _ = self.bert_spc(
            text_target_bert_indices, text_target_bert_segments_ids
        )

        bert_local_out, _, _ = self.bert_local(text_local_indices)

        if self.opt.local_context_focus == "cdm":
            masked_local_text_vec = self.feature_dynamic_mask(
                text_local_indices, aspect_indices, syntax_hop_distance_to_target
            )
            bert_local_out = torch.mul(bert_local_out, masked_local_text_vec)

        elif self.opt.local_context_focus == "cdw":
            weighted_text_local_features = self.feature_dynamic_weighted(
                text_local_indices, aspect_indices, syntax_hop_distance_to_target
            )
            bert_local_out = torch.mul(bert_local_out, weighted_text_local_features)

        out_cat = torch.cat((bert_local_out, bert_spc_out), dim=-1)
        mean_pool = self.mean_pooling_double(out_cat)
        self_attention_out = self.bert_SA(mean_pool)
        pooled_out = self.bert_pooler(self_attention_out)

        dense_out = self.dense(pooled_out)
        return dense_out



import torch

from NewsSentiment.knowledge.knowledgeutils import find_key_original_or_lc
from NewsSentiment.knowledge.nrcemolex.converter import PATH_DICT_NRC_EMOLEX, EMOTION2INDEX
from NewsSentiment.diskdict import DiskDict

__ddict_emolex = DiskDict(PATH_DICT_NRC_EMOLEX)
__ddict_emolex_keys_lower = {k.lower(): v for k, v in __ddict_emolex.items()}
__num_emotions = 10
assert len(__ddict_emolex) == 6468
assert len(EMOTION2INDEX) == __num_emotions


def get_num_nrc_emotions():
    return __num_emotions


def get_nrc_emotions_as_tensor(term: str):
    emotions = find_key_original_or_lc(__ddict_emolex, __ddict_emolex_keys_lower, term)
    tensor_emotions = torch.zeros(__num_emotions, dtype=torch.long)
    for emotion_index in emotions:
        tensor_emotions[emotion_index] = 1
    return tensor_emotions



from NewsSentiment.consts import BERT_BASE_UNCASED

combinations_default_0 = {
    "own_model_name": [
        # baselines: single
        "notargetclsbert",
        "lcf_bert",
        "lcf_bert2",
        "lcfs_bert",
        "lcft_bert",
        "aen_bert",
        "spc_bert",
        "tdbert",
        "tdbert-qa-mul",
        "tdbert-qa-con",
        # own models: single
        "tdbertlikesingle",
        "lcfst_bert",
        "grutsc",
        # own models: multi
        "tdbertlikemulti",
        # "tdbertlikemulti_dense",
        "seq2seq",
        "seq2seq_withouttargetmask",
        "contrasting",
        # baselines
        # "random_single",
        # "random_multi",
    ],
    "optimizer": ["adam"],
    "initializer": ["xavier_uniform_"],
    "learning_rate": ["2e-5", "3e-5", "5e-5"],
    "batch_size": [
        "16",
        # "32",
    ],  # ['16', '32'],
    "balancing": ["None"],  # ['None', 'lossweighting', 'oversampling'],
    "devmode": ["False"],
    "num_epoch": ["2", "3", "4"],
    "loss": [
        "crossentropy",
        "crossentropy_lsr",
        "sequence",
        "crossentropy_crossweight",
    ],
    # "spc_lm_representation_distilbert": ["mean_last"],
    # ['sum_last', 'sum_last_four', 'sum_last_two', 'sum_all', 'mean_last', 'mean_last_four', 'mean_last_two', 'mean_all'],
    # "spc_lm_representation": ["pooler_output"],
    # ['pooler_output', 'sum_last', 'sum_last_four', 'sum_last_two', 'sum_all', 'mean_last', 'mean_last_four', 'mean_last_two', 'mean_all'],
    # "spc_input_order": ["text_target"],  # 'target_text',
    # "aen_lm_representation": ["last"],
    # ['last', 'sum_last_four', 'sum_last_two', 'sum_all', 'mean_last_four'],  # 'mean_last_two', 'mean_all'],
    "eval_only_after_last_epoch": ["True"],
    "local_context_focus": ["cdm", "cdw"],
    "SRD": ["3", "4", "5"],
    "pretrained_model_name": ["default"],
    # ['default', 'bert_news_ccnc_10mio_3ep', 'laptops_and_restaurants_2mio_ep15', 'laptops_1mio_ep30', 'restaurants_10mio_ep3'],
    "state_dict": ["None"],
    # ['None', 'lcf_bert_acl14twitter_val_recall_avg_0.7349_epoch3', 'lcf_bert_semeval14laptops_val_recall_avg_0.7853_epoch3', 'lcf_bert_semeval14restaurants_val_recall_avg_0.7672_epoch2', 'lcf_bert_newstsc_val_recall_avg_0.5954_epoch3'],
    "single_targets": [
        "True"
    ],  # using conditions in controller.py, we have single_targets only for single target models
    "multi_targets": [
        "True"
    ],  # using conditions in controller.py, we have multi_targets only for multi target models
    "targetclasses": [
        "newsmtsc3",
        #"newsmtsc3strong",
        #"newsmtsc3weak",
    ],
    "knowledgesources": [
        "nrc_emotions", "mpqa_subjectivity", "bingliu_opinion", "liwc",
        "nrc_emotions mpqa_subjectivity", "nrc_emotions liwc",
        "nrc_emotions bingliu_opinion", "mpqa_subjectivity bingliu_opinion",
        "mpqa_subjectivity liwc", "bingliu_opinion liwc",
        "nrc_emotions mpqa_subjectivity bingliu_opinion",
        "nrc_emotions mpqa_subjectivity liwc",
        "nrc_emotions liwc bingliu_opinion",
        "liwc mpqa_subjectivity bingliu_opinion",
        "nrc_emotions mpqa_subjectivity bingliu_opinion liwc",
        "zeros",
    ],
    "is_use_natural_target_phrase_for_spc": [
        "True",
        "False"
    ],
    "default_lm": [
        BERT_BASE_UNCASED,
    ],
    "coref_mode_in_training": [
        "ignore",
        "in_targetmask",
        "additional_examples"
    ],
}



BERT_BASE_UNCASED = "bert-base-uncased"
ROBERTA_BASE = "roberta-base"
XLNET_BASE_CASED = "xlnet-base-cased"
ALBERT_BASE = "albert-base-v2"
ALBERT_LARGE = "albert-large-v2"
ALBERT_XLARGE = "albert-xlarge-v2"
ALBERT_XXLARGE = "albert-xxlarge-v2"
__DEFAULT_LM = None


def set_default_lm(new_name: str):
    global __DEFAULT_LM
    __DEFAULT_LM = new_name


def get_default_lm():
    return __DEFAULT_LM


FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS = "text_ids_with_special_tokens"
FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK = (
    FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS + "_target_mask"
)
FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_SELECTED_KNOWLEDGE_SOURCES = (
    FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS + "_selectedknowledgesources"
)
FIELD_IS_OVERFLOW = "is_overflow"
FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS = (
    "text_then_target_ids_with_special_tokens"
)
# we used to have text-then-target target mask here, but won't use it,
# since it would be identical to the text target mask (since we only
# want to mark the target within the text, but not in the 2nd target
# component)
# FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK = (
#    FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS + "_target_mask"
# )
# same for knowledge sources
# FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS_KNOWLEDGE_SOURCES = (
#         FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS + "_selectedknowledgesources"
# )
FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS_SEGMENT_IDS = (
    FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS + "_segment_ids"
)
FIELD_TARGET_IDS_WITH_SPECIAL_TOKENS = "target_ids_with_special_tokens"
FIELD_SYNTAX_HOP_DISTANCE_TO_TARGET = "syntax_hop_distance_to_target"
FIELD_SYNTAX_DEPENDENCY_MATRIX = "syntax_dependency_matrix"



import os
from collections import defaultdict, Counter
from pathlib import Path
import csv

from tqdm import tqdm

from NewsSentiment.fxlogger import get_logger
from NewsSentiment.diskdict import DiskDict

POLARITY2INDEX = {
    "positive": 2,
    "neutral": 1,
    "negative": 0,
}
THIS_DIR = Path(os.path.dirname(os.path.realpath(__file__)))
PATH_DICT_MPQA_SUBJECTIVITY = THIS_DIR / "subjclueslen1-HLTEMNLP05.tff.ddict"

logger = get_logger()


def get_value(entry_key_value: str):
    return entry_key_value.split("=")[1]


def convert_txt_to_dict():
    path_dict = THIS_DIR / "subjclueslen1-HLTEMNLP05.tff"

    term2polarity = defaultdict(set)
    polarity_counter = Counter()
    with open(path_dict, "r") as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=" ")
        for line in csv_reader:
            entry_type, entry_length, entry_word, entry_pos, entry_is_stemmed, entry_polarity = line
            entry_type = get_value(entry_type)
            entry_length = get_value(entry_length)
            entry_word = get_value(entry_word)
            entry_pos = get_value(entry_pos)
            entry_is_stemmed = get_value(entry_is_stemmed)
            entry_polarity = get_value(entry_polarity)

            assert " " not in entry_word, f"cannot handle spaces in word"
            assert entry_polarity in POLARITY2INDEX.keys() or entry_polarity == "both", f"polarity label not known: {entry_polarity} for {entry_word}"

            if entry_polarity == "both":
                polarities = [POLARITY2INDEX["positive"], POLARITY2INDEX["negative"]]
            else:
                polarities = (POLARITY2INDEX[entry_polarity],)

            for polarity in polarities:
                term2polarity[entry_word].add(polarity)
                polarity_counter[polarity] += 1

    logger.info("read %s terms", len(term2polarity))
    logger.info("polarity count:\n%s", polarity_counter.most_common())

    logger.info("saving to %s...", PATH_DICT_MPQA_SUBJECTIVITY)
    ddict_emolex = DiskDict(PATH_DICT_MPQA_SUBJECTIVITY)
    ddict_emolex.update(term2polarity)
    ddict_emolex.sync_to_disk()
    logger.info("done")


if __name__ == "__main__":
    convert_txt_to_dict()



from collections import Counter
from statistics import mean

import jsonlines
import numpy as np
from sklearn import metrics

from NewsSentiment.SentimentClasses import SentimentClasses
from NewsSentiment.dataset import FXDataset
from NewsSentiment.fxlogger import get_logger


class Evaluator:
    def __init__(self, sorted_expected_label_values, polarity_associations, snem_name):
        self.logger = get_logger()
        self.polarity_associations = polarity_associations
        self.pos_label_value = polarity_associations["positive"]
        self.neg_label_value = polarity_associations["negative"]
        self.sorted_expected_label_values = sorted_expected_label_values
        self.pos_label_index = self.sorted_expected_label_values.index(
            self.pos_label_value
        )
        self.neg_label_index = self.sorted_expected_label_values.index(
            self.neg_label_value
        )
        self.snem_name = snem_name

    def mean_from_all_statistics(self, all_test_stats):
        # for Counters, we do not take the mean
        mean_test_stats = {}
        number_stats = len(all_test_stats)

        for key in all_test_stats[0]:
            value_type = type(all_test_stats[0][key])

            if value_type in [float, np.float64, np.float32]:
                aggr_val = 0.0
                for test_stat in all_test_stats:
                    aggr_val += test_stat[key]

                mean_test_stats[key] = aggr_val / number_stats

            elif value_type == Counter:
                aggr_val = Counter()
                for test_stat in all_test_stats:
                    aggr_val += test_stat[key]
                mean_test_stats[key] = aggr_val

        return mean_test_stats

    def calc_statistics(self, y_true, y_pred):
        """
        Calculates performance statistics by comparing k-dimensional Tensors y_true and
        y_pred. Both y_true and y_pred's shape have to be(batchsize, maxtargetsperexample)
        :param y_true:
        :param y_pred:
        :return:
        """
        assert y_true.shape == y_pred.shape, "different shapes"
        assert y_true.shape[1] == FXDataset.NUM_MAX_TARGETS_PER_ITEM

        # for now, the following doesn't keep track of which prediction or true answer
        # belongs to which examples. for other measures, this might be necessary, though
        # in that case, the code would have to be changed, e.g., by keeping the original
        # dimensions. for now, we just unpack the Tensors to so that the former
        # evaluation logic will work on them. practically, this treats each non-fillup
        # target as an example
        y_true = y_true.view(-1)
        y_pred = y_pred.view(-1)

        # in both tensors, keep only those scalars that are non-fill in y_true
        non_fillup_mask = y_true != SentimentClasses.FILLUP_POLARITY_VALUE
        y_true = y_true[non_fillup_mask]
        y_pred = y_pred[non_fillup_mask]

        # this is just the previous, single-target evaluation logic
        y_true_list = y_true.tolist()
        y_pred_list = y_pred.tolist()
        y_true_count = Counter(y_true_list)
        y_pred_count = Counter(y_pred_list)

        f1_macro = metrics.f1_score(
            y_true, y_pred, labels=self.sorted_expected_label_values, average="macro"
        )
        f1_of_classes = metrics.f1_score(
            y_true, y_pred, labels=self.sorted_expected_label_values, average=None
        )
        f1_posneg = (
            f1_of_classes[self.pos_label_index] + f1_of_classes[self.neg_label_index]
        ) / 2.0
        confusion_matrix = metrics.confusion_matrix(
            y_true, y_pred, labels=self.sorted_expected_label_values
        )
        recalls_of_classes = metrics.recall_score(
            y_true, y_pred, labels=self.sorted_expected_label_values, average=None
        )
        recall_avg = mean(recalls_of_classes)
        recall_macro = metrics.recall_score(
            y_true, y_pred, labels=self.sorted_expected_label_values, average="macro"
        )
        precision_macro = metrics.precision_score(
            y_true, y_pred, labels=self.sorted_expected_label_values, average="macro"
        )
        accuracy = metrics.accuracy_score(y_true, y_pred)

        return {
            "f1_macro": f1_macro,
            "confusion_matrix": confusion_matrix,
            "recalls_of_classes": recalls_of_classes,
            "recall_avg": recall_avg,
            "recall_macro": recall_macro,
            "precision_macro": precision_macro,
            "accuracy": accuracy,
            "f1_posneg": f1_posneg,
            "y_true_count": y_true_count,
            "y_pred_count": y_pred_count,
        }

    def print_stats(self, stats, description):
        self.logger.info(description)
        self.logger.info("{}: {})".format(self.snem_name, stats[self.snem_name]))
        self.logger.info(
            "y_true distribution: {}".format(sorted(stats["y_true_count"].items()))
        )
        self.logger.info(
            "y_pred distribution: {}".format(sorted(stats["y_pred_count"].items()))
        )
        self.logger.info(
            "> recall_avg: {:.4f}, f1_posneg: {:.4f}, acc: {:.4f}, f1_macro: {:.4f}".format(
                stats["recall_avg"],
                stats["f1_posneg"],
                stats["accuracy"],
                stats["f1_macro"],
            )
        )

    def write_error_table(self, y_true, y_pred, texts_list, filepath):
        y_true_list = y_true.tolist()
        y_pred_list = y_pred.tolist()

        with jsonlines.open(filepath, "w") as writer:
            for true_label, pred_label, text in zip(
                y_true_list, y_pred_list, texts_list
            ):
                writer.write(
                    {"true_label": true_label, "pred_label": pred_label, "text": text}
                )
